# -*- Mode: python; tab-width: 4; indent-tabs-mode:nil; coding:utf-8 -*-
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4 fileencoding=utf-8
#
# MDAnalysis --- https://www.mdanalysis.org
# Copyright (c) 2006-2017 The MDAnalysis Development Team and contributors
# (see the file AUTHORS for the full list of names)
#
# Released under the Lesser GNU Public Licence, v2.1 or any higher version
#
# Please cite your use of MDAnalysis in published work:
#
# R. J. Gowers, M. Linke, J. Barnoud, T. J. E. Reddy, M. N. Melo, S. L. Seyler,
# D. L. Dotson, J. Domanski, S. Buchoux, I. M. Kenney, and O. Beckstein.
# MDAnalysis: A Python package for the rapid analysis of molecular dynamics
# simulations. In S. Benthall and S. Rostrup editors, Proceedings of the 15th
# Python in Science Conference, pages 102-109, Austin, TX, 2016. SciPy.
# doi: 10.25080/majora-629e541a-00e
#
# N. Michaud-Agrawal, E. J. Denning, T. B. Woolf, and O. Beckstein.
# MDAnalysis: A Toolkit for the Analysis of Molecular Dynamics Simulations.
# J. Comput. Chem. 32 (2011), 2319--2327, doi:10.1002/jcc.21787
#

"""
Location of data files for the MDAnalysis unit tests
====================================================

Real MD simulation data are stored in the ``data/`` sub
directory. Use as ::

  from MDAnalysis.tests.datafiles import *

Note that the files are actually located in a separate package,
:mod:`MDAnalysisTestData` from where they are initially imported as ::

 from MDAnalysisTestData.datafiles import *
"""

__all__ = [
    "PSF",
    "DCD",
    "CRD",  # CHARMM (AdK example, DIMS trajectory from JMB 2009 paper)
    "DCD2",  # CHARMM (AdK example, DIMS trajectory from PLOS Comput Biol paper)
    "PSF_notop",
    "PSF_BAD",  # Same as PSF but no bonds etc, malformed version of previous
    "DCD_empty",
    "PSF_TRICLINIC",
    "DCD_TRICLINIC",  # CHARMM c36 new unitcell, NPT 125 TIP3P (box vectors, see Issue 187 for details)
    "PSF_NAMD",
    "PDB_NAMD",  # NAMD
    "PSF_NAMD_TRICLINIC",
    "DCD_NAMD_TRICLINIC",  # NAMD, triclinic unitcell (Issue 187)
    "PSF_NAMD_GBIS",
    "DCD_NAMD_GBIS",  # NAMD, implicit solvent, 100 steps,  #1819
    "PSF_nosegid",  # psf without a segid, Issue 121
    "PSF_cmap",  # ala3 PSF from ParmEd test files with cmap
    "PSF_inscode",  # PSF file with insertion codes
    "PDB_small",  # PDB
    "PDB_xsmall",  # The first 10 residues of PDB_small
    "PDB_closed",
    "PDB_multiframe",
    "PDB_helix",
    "PDB_conect",
    "PDB_conect2TER",  # Conect record to a TER entry (Issue 936)
    "PDB_singleconect",  # Conect record with one entry (Issue 937)
    "PDB_icodes",  # stripped down version of 1osm, has icodes!
    "PDB_varying",  # varying occupancies and tempfactors
    "XPDB_small",
    "PDB_full",  # PDB 4E43 (full HEADER, TITLE, COMPND, REMARK, altloc)
    "ALIGN",  # Various way to align atom names in PDB files
    "RNA_PSF",
    "RNA_PDB",  # nucleic acid (PDB 1K5I in CHARMM36m)
    "INC_PDB",  # incomplete PDB file (Issue #396)
    # for testing cryst before/after model headers
    "PDB_cm",
    "PDB_cm_bz2",
    "PDB_cm_gz",
    "PDB_mc",
    "PDB_mc_bz2",
    "PDB_mc_gz",
    "PDB_chainidnewres",  # Issue 1110
    "PDB_sameresid_diffresname",  # Case where two residues share the same resid
    "PDB_chainidrepeat",  # Issue #1107
    "PDB",
    "GRO",
    "XTC",
    "TRR",
    "TPR",
    "GRO_velocity",  # Gromacs (AdK)
    "GRO_incomplete_vels",
    "COORDINATES_GRO_BZ2",
    "GRO_large",  # atom number truncation at > 100,000 particles, Issue 550
    "GRO_residwrap",  # resids wrapping because of 5 digit field (Issue #728)
    "GRO_residwrap_0base",  # corner case of #728 with resid=0 for first atom
    "GRO_sameresid_diffresname",  # Case where two residues share the same resid
    "PDB_xvf",
    "TPR_xvf",
    "TRR_xvf",  # Gromacs coords/veloc/forces (cobrotoxin, OPLS-AA, Gromacs 4.5.5 tpr)
    "TPR_xvf_2024_4",
    "H5MD_xvf",  # TPR_xvf + TRR_xvf converted to h5md format
    "H5MD_energy",  # H5MD trajectory with observables/atoms/energy
    "H5MD_malformed",  # H5MD trajectory with malformed observable group
    "XVG_BZ2",  # Compressed xvg file about cobrotoxin
    "PDB_xlserial",
    "TPR400",
    "TPR402",
    "TPR403",
    "TPR404",
    "TPR405",
    "TPR406",
    "TPR407",
    "TPR450",
    "TPR451",
    "TPR452",
    "TPR453",
    "TPR454",
    "TPR455",
    "TPR455Double",
    "TPR460",
    "TPR461",
    "TPR502",
    "TPR504",
    "TPR505",
    "TPR510",
    "TPR2016",
    "TPR2018",
    "TPR2019B3",
    "TPR2020B2",
    "TPR2020",
    "TPR2020Double",
    "TPR2021",
    "TPR2021Double",
    "TPR2022RC1",
    "TPR2023",
    "TPR2024",
    "TPR2024_4",
    "TPR2025_0",
    "TPR510_bonded",
    "TPR2016_bonded",
    "TPR2018_bonded",
    "TPR2019B3_bonded",
    "TPR2020B2_bonded",
    "TPR2020_bonded",
    "TPR2020_double_bonded",
    "TPR2021_bonded",
    "TPR2021_double_bonded",
    "TPR2022RC1_bonded",
    "TPR334_bonded",
    "TPR2023_bonded",
    "TPR2024_bonded",
    "TPR2024_4_bonded",
    "TPR2025_0_bonded",
    "TPR_EXTRA_2021",
    "TPR_EXTRA_2020",
    "TPR_EXTRA_2018",
    "TPR_EXTRA_2016",
    "TPR_EXTRA_407",
    "TPR_EXTRA_2022RC1",
    "TPR_EXTRA_2023",
    "TPR_EXTRA_2024",
    "TPR_EXTRA_2024_4",
    "TPR_EXTRA_2025_0",
    "TPR_NNPOT_2025_0",
    "PDB_sub_sol",
    "PDB_sub_dry",  # TRRReader sub selection
    "TRR_sub_sol",
    "XTC_sub_sol",
    "XYZ",
    "XYZ_psf",
    "XYZ_bz2",
    "XYZ_mini",
    "XYZ_five",  # 3 and 5 atoms xyzs for an easy topology
    "TXYZ",
    "ARC",
    "ARC_PBC",  # Tinker files
    "PRM",
    "PRM_chainid_bz2",
    "TRJ",
    "TRJ_bz2",  # Amber (no periodic box)
    "INPCRD",
    "PRMpbc",
    "TRJpbc_bz2",  # Amber (periodic box)
    "PRM7",
    "NCDFtruncoct",  # Amber (cpptrj test trajectory, see Issue 488)
    "PRM12",
    "TRJ12_bz2",  # Amber (v12 format, Issue 100)
    "PRMncdf",
    "TRJncdf",
    "NCDF",  # Amber (netcdf)
    "PFncdf_Top",
    "PFncdf_Trj",  # Amber ncdf with Positions and Forces
    "CPPTRAJ_TRAJ_TOP",
    "CPPTRAJ_TRAJ",  # Amber ncdf extracted from CPPTRAJ without time variable
    "PRMcs",  # Amber (format, Issue 1331)
    "PRMNCRST",  # Amber ncrst with positions/forces/velocities
    "PRM_NCBOX",
    "TRJ_NCBOX",  # Amber parm7 + nc w/ pos/forces/vels/box
    "PRMNEGATIVE",  # Amber negative ATOMIC_NUMBER (Issue 2306)
    "PRMErr1",  # Amber TOP files to check raised errors
    "PRMErr2",
    "PRMErr3",
    "PRMErr4",
    "PRMErr5",
    "PRM_UreyBradley",  # prmtop from ParmEd test files with Urey-Bradley angles
    "PRM7_ala2",
    "RST7_ala2",  # prmtop and rst files from ParmEd example files
    "PRM19SBOPC",  #  prmtop w/ ff19SB CMAP terms and OPC water (Issue #2449)
    "PQR",  # PQR v1
    "PQR_icodes",  # PQR v2 with icodes
    "PDBQT_input",  # PDBQT
    "PDBQT_querypdb",
    "PDBQT_tyrosol",
    "PDB_multipole",
    "FASTA",  # sequence alignment, Issue 112 + 113
    "HELANAL_BENDING_MATRIX",  # HELANAL test (from PSF+DCD (AdK) helix 8)
    "HELANAL_BENDING_MATRIX_SUBSET",  # As above, slice of frames 10 to 79
    "PDB_HOLE",  # gramicidin A
    "MULTIPDB_HOLE",  # gramicidin A, normal mode 7 from ElNemo
    "DMS",
    "DMS_DOMAINS",  # ADK closed with multiple segids
    "DMS_NO_SEGID",  # ADK closed with no segids or chains
    "CONECT",  # HIV Reverse Transcriptase with inhibitor
    "CONECT_ERROR",  # PDB file with corrupt CONECT
    "TRZ",
    "TRZ_psf",
    "TRIC",
    "XTC_multi_frame",
    "TRR_multi_frame",
    "TNG_traj",  # TNG trajectory from GROMACS physical validation testsuite, longish trajectory
    "TNG_traj_gro",  # topology for argon_npt_compressed_traj
    "TNG_traj_uneven_blocks",  # TNG trajectory with pos and vel deposited on different strides
    "TNG_traj_vels_forces",  # similar to above but with velocities and forces
    "merge_protein",
    "merge_ligand",
    "merge_water",
    "mol2_molecules",
    "mol2_molecule",
    "mol2_broken_molecule",
    "mol2_zinc",
    "mol2_comments_header",
    "mol2_ligand",
    "mol2_sodium_ion",
    "capping_input",
    "capping_output",
    "capping_ace",
    "capping_nma",
    "contacts_villin_folded",
    "contacts_villin_unfolded",
    "contacts_file",
    "LAMMPSdata",
    "trz4data",
    "LAMMPSdata_mini",
    "LAMMPSdata2",
    "LAMMPSdcd2",
    "LAMMPScnt",
    "LAMMPScnt2",  # triclinic box
    "LAMMPShyd",
    "LAMMPShyd2",
    "LAMMPSdata_many_bonds",
    "LAMMPSdata_deletedatoms",  # with deleted atoms
    "LAMMPSdata_triclinic",  # lammpsdata file to test triclinic dimension parsing, albite with most atoms deleted
    "LAMMPSdata_PairIJ",  # lammps datafile with a PairIJ Coeffs section
    "LAMMPSdata_additional_columns",  # structure for the additional column lammpstrj
    "LAMMPSDUMP",
    "LAMMPSDUMP_long",  # lammpsdump file with a few zeros sprinkled in the first column first frame
    "LAMMPSDUMP_allinfo",  # lammpsdump file with resids, masses, charges and element symbols
    "LAMMPSDUMP_nomass_elemx",  # lammps dump file with no masses, but with element symbols and one symbol 'X'
    "LAMMPSDUMP_allcoords",  # lammpsdump file with all coordinate conventions (x,xs,xu,xsu) present, from LAMMPS rdf example
    "LAMMPSDUMP_nocoords",  # lammpsdump file with no coordinates
    "LAMMPSDUMP_triclinic",  # lammpsdump file to test triclinic dimension parsing, albite with most atoms deleted
    "LAMMPSDUMP_image_vf",  # Lammps dump file with image flags, velocities, and forces.
    "LAMMPS_image_vf",  # Lammps data file to go with LAMMPSDUMP_image_vf
    "LAMMPSDUMP_chain1",  # Lammps dump file with chain reader
    "LAMMPSDUMP_chain2",  # Lammps dump file with chain reader
    "LAMMPS_chain",  # Lammps data file with chain reader
    "LAMMPSDUMP_additional_columns",  # lammpsdump file with additional data (an additional charge column)
    "LAMMPSDUMP_non_linear",  # lammpsdump file to test calculating non-linear msd method
    "unordered_res",  # pdb file with resids non sequential
    "GMS_ASYMOPT",  # GAMESS C1  optimization
    "GMS_SYMOPT",  # GAMESS D4h optimization
    "GMS_ASYMSURF",  # GAMESS C1  surface
    "two_water_gro",
    "two_water_gro_nonames",  # for bond guessing, 2 water molecules, one with weird names
    "two_water_gro_multiframe",
    "two_water_gro_widebox",  # Issue #548
    "DLP_CONFIG",
    "DLP_CONFIG_order",
    "DLP_CONFIG_minimal",  # dl_poly 4 config file
    "DLP_HISTORY",
    "DLP_HISTORY_order",
    "DLP_HISTORY_minimal",  # dl_poly 4 history file
    "DLP_HISTORY_minimal_cell",  # dl_poly 4 history file with cell parameters
    "DLP_HISTORY_classic",  # dl_poly classic history file
    "waterPSF",
    "waterDCD",
    "rmsfArray",
    "HoomdXMLdata",
    "Make_Whole",  # for testing the function lib.mdamath.make_whole, has 9 atoms
    "fullerene",  # for make_whole, a nice friendly C60 with bonds
    "Plength",
    "COORDINATES_XYZ",
    "COORDINATES_XYZ_BZ2",
    "COORDINATES_GRO",
    "COORDINATES_GRO_INCOMPLETE_VELOCITY",
    "Martini_membrane_gro",  # for testing the leaflet finder
    "COORDINATES_XTC",
    "COORDINATES_TRR",
    "COORDINATES_TNG",
    "COORDINATES_H5MD",
    "COORDINATES_DCD",
    "COORDINATES_TOPOLOGY",
    "NUCLsel",
    "GRO_empty_atom",
    "GRO_missing_atomname",  # for testing GROParser exception raise
    "ENT",  # for testing ENT file extension
    "RANDOM_WALK",
    "RANDOM_WALK_TOPO",  # garbage topology to go along with XTC positions above
    "AUX_XVG",
    "XVG_BAD_NCOL",  # for testing .xvg auxiliary reader
    "AUX_XVG_LOWF",
    "AUX_XVG_HIGHF",
    "AUX_EDR",
    "AUX_EDR_TPR",
    "AUX_EDR_XTC",
    "AUX_EDR_RAW",
    "AUX_EDR_SINGLE_FRAME",  # for testing .edr auxiliary reader
    "MMTF",
    "MMTF_gz",
    "MMTF_skinny",  # skinny - some optional fields stripped out
    "MMTF_skinny2",
    "ALIGN_BOUND",  # two component bound system
    "ALIGN_UNBOUND",  # two component unbound system
    "legacy_DCD_ADK_coords",  # frames 5 and 29 read in for adk_dims.dcd using legacy DCD reader
    "legacy_DCD_NAMD_coords",  # frame 0 read in for SiN_tric_namd.dcd using legacy DCD reader
    "legacy_DCD_c36_coords",  # frames 1 and 4 read in for tip125_tric_C36.dcd using legacy DCD reader
    "GSD",
    "GSD_bonds",
    "GSD_long",
    "TRC_PDB_VAC",
    "TRC_TRAJ1_VAC",
    "TRC_TRAJ1_VAC_WHITESPACE",  # contains inconsistent trailing whitespace in POSITIONRED
    "TRC_TRAJ1_VAC_MISSING_POS",  # contains one missing position in the last frame
    "TRC_TRAJ1_VAC_EXTRA_POS",  # contains one extra position in the last frame
    "TRC_TRAJ2_VAC",  # 2x 3 frames of vacuum trajectory from GROMOS11 tutorial
    "TRC_CLUSTER_VAC",  # three frames without TIMESTEP and GENBOX block but with unsupported POSITION block
    "TRC_TRICLINIC_SOLV",
    "TRC_TRUNCOCT_VAC",
    "TRC_GENBOX_ORIGIN",
    "TRC_GENBOX_EULER",
    "TRC_EMPTY",  # Empty file containing only one space
    "TRC_PDB_SOLV",
    "TRC_TRAJ_SOLV",  # 2 frames of solvated trajectory from GROMOS11 tutorial
    "GRO_MEMPROT",
    "XTC_MEMPROT",  # YiiP transporter in POPE:POPG lipids with Na+, Cl-, Zn2+ dummy model without water
    "DihedralArray",
    "DihedralsArray",  # time series of single dihedral
    "RamaArray",
    "GLYRamaArray",  # time series of phi/psi angles
    "JaninArray",
    "LYSJaninArray",  # time series of chi1/chi2 angles
    "PDB_rama",
    "PDB_janin",  # for testing failures of Ramachandran and Janin classes
    "BATArray",  # time series of bond-angle-torsion coordinates array from Molecule_comments_header.mol2
    # DOS line endings
    "WIN_PDB_multiframe",
    "WIN_DLP_HISTORY",
    "WIN_TRJ",
    "WIN_LAMMPSDUMP",
    "WIN_ARC",
    "GRO_huge_box",  # for testing gro parser with hige box sizes
    "ITP",  # for GROMACS generated itps
    "ITP_nomass",  # for ATB generated itps
    "ITP_atomtypes",  # atom definitions to check atomtyes section parsing
    "ITP_charges",  # atom definitions to test custom particle charge parsing.
    "NAMDBIN",  # for NAMD generated binary file
    "ITP_edited",  # to check different directives are read properly
    "ITP_tip5p",  # tip5p water from opls-aa, edited with additional keywords
    "ITP_spce",  # spce water from gromos54a7, edited with additional keywords,
    "GMX_TOP",  # 2 ala10 chains + 3 spc water
    "GMX_DIR",  # GROMACS directory
    "GMX_TOP_BAD",  # file with an #include that doesn't exist
    "ITP_no_endif",  # file missing an #endif
    "PDB_CRYOEM_BOX",  # Issue 2599, Issue #2679, PR #2685
    "PDB_CHECK_RIGHTHAND_PA",  # for testing right handedness of principal_axes
    "MMTF_NOCRYST",  # File with meaningless CRYST1 record (Issue #2679, PR #2685)
    "FHIAIMS",  # to test FHIAIMS coordinate files
    "SDF_molecule",  # MDL SDFile for rdkit
    "PDBX",  # PDBxfile
    "PDB_elements",  # PDB file with elements
    "PDB_charges",  # PDB file with formal charges
    "SURFACE_PDB",  # 111 FCC lattice topology for NSGrid bug #2345
    "SURFACE_TRR",  # full precision coordinates for NSGrid bug #2345
    "DSSP",  # DSSP test suite
]

from importlib import resources
import MDAnalysisTests.data

_data_ref = resources.files("MDAnalysisTests.data")

WIN_PDB_multiframe = (
    _data_ref / "windows/WIN_nmr_neopetrosiamide.pdb"
).as_posix()
WIN_DLP_HISTORY = (_data_ref / "windows/WIN_HISTORY").as_posix()
WIN_TRJ = (_data_ref / "windows/WIN_ache.mdcrd").as_posix()
WIN_ARC = (_data_ref / "windows/WIN_test.arc").as_posix()
WIN_LAMMPSDUMP = (_data_ref / "windows/WIN_wat.lammpstrj").as_posix()

legacy_DCD_NAMD_coords = (_data_ref / "legacy_DCD_NAMD_coords.npy").as_posix()
legacy_DCD_ADK_coords = (_data_ref / "legacy_DCD_adk_coords.npy").as_posix()
legacy_DCD_c36_coords = (_data_ref / "legacy_DCD_c36_coords.npy").as_posix()
AUX_XVG_LOWF = (_data_ref / "test_lowf.xvg").as_posix()
AUX_XVG_HIGHF = (_data_ref / "test_highf.xvg").as_posix()
XVG_BAD_NCOL = (_data_ref / "bad_num_col.xvg").as_posix()
AUX_XVG = (_data_ref / "test.xvg").as_posix()
AUX_EDR = (_data_ref / "test.edr").as_posix()
AUX_EDR_RAW = (_data_ref / "aux_edr_raw.txt").as_posix()
AUX_EDR_TPR = (_data_ref / "aux_edr.tpr").as_posix()
AUX_EDR_XTC = (_data_ref / "aux_edr.xtc").as_posix()
AUX_EDR_SINGLE_FRAME = (_data_ref / "single_frame.edr").as_posix()
ENT = (_data_ref / "testENT.ent").as_posix()
GRO_missing_atomname = (_data_ref / "missing_atomname.gro").as_posix()
GRO_empty_atom = (_data_ref / "empty_atom.gro").as_posix()
GRO_huge_box = (_data_ref / "huge_box.gro").as_posix()

COORDINATES_GRO = (_data_ref / "coordinates/test.gro").as_posix()
COORDINATES_GRO_INCOMPLETE_VELOCITY = (
    _data_ref / "coordinates/test_incomplete_vel.gro"
).as_posix()
COORDINATES_GRO_BZ2 = (_data_ref / "coordinates/test.gro.bz2").as_posix()
COORDINATES_XYZ = (_data_ref / "coordinates/test.xyz").as_posix()
COORDINATES_XYZ_BZ2 = (_data_ref / "coordinates/test.xyz.bz2").as_posix()
COORDINATES_XTC = (_data_ref / "coordinates/test.xtc").as_posix()
COORDINATES_TRR = (_data_ref / "coordinates/test.trr").as_posix()
COORDINATES_TNG = (_data_ref / "coordinates/test.tng").as_posix()
COORDINATES_H5MD = (_data_ref / "coordinates/test.h5md").as_posix()
COORDINATES_DCD = (_data_ref / "coordinates/test.dcd").as_posix()
COORDINATES_TOPOLOGY = (_data_ref / "coordinates/test_topology.pdb").as_posix()

PSF = (_data_ref / "adk.psf").as_posix()
PSF_notop = (_data_ref / "adk_notop.psf").as_posix()
PSF_BAD = (_data_ref / "adk_notop_BAD.psf").as_posix()
DCD = (_data_ref / "adk_dims.dcd").as_posix()
DCD_empty = (_data_ref / "empty.dcd").as_posix()
CRD = (_data_ref / "adk_open.crd").as_posix()
PSF_TRICLINIC = (_data_ref / "tip125_tric_C36.psf").as_posix()
DCD_TRICLINIC = (_data_ref / "tip125_tric_C36.dcd").as_posix()
DCD2 = (_data_ref / "adk_dims2.dcd").as_posix()

PSF_NAMD = (_data_ref / "namd_cgenff.psf").as_posix()
PDB_NAMD = (_data_ref / "namd_cgenff.pdb").as_posix()
PDB_multipole = (
    _data_ref / "water_methane_acetic-acid_ammonia.pdb"
).as_posix()
PSF_NAMD_TRICLINIC = (_data_ref / "SiN_tric_namd.psf").as_posix()
DCD_NAMD_TRICLINIC = (_data_ref / "SiN_tric_namd.dcd").as_posix()
PSF_NAMD_GBIS = (_data_ref / "adk_closed_NAMD.psf").as_posix()
DCD_NAMD_GBIS = (_data_ref / "adk_gbis_tmd-fast1_NAMD.dcd").as_posix()

PSF_nosegid = (_data_ref / "nosegid.psf").as_posix()

PSF_cmap = (_data_ref / "parmed_ala3.psf").as_posix()

PSF_inscode = (_data_ref / "1a2c_ins_code.psf").as_posix()

PDB_varying = (_data_ref / "varying_occ_tmp.pdb").as_posix()
PDB_small = (_data_ref / "adk_open.pdb").as_posix()
PDB_xsmall = (_data_ref / "adk_open_10res.pdb").as_posix()
PDB_closed = (_data_ref / "adk_closed.pdb").as_posix()

ALIGN = (_data_ref / "align.pdb").as_posix()
RNA_PSF = (_data_ref / "analysis/1k5i_c36.psf.gz").as_posix()
RNA_PDB = (_data_ref / "analysis/1k5i_c36.pdb.gz").as_posix()
INC_PDB = (_data_ref / "incomplete.pdb").as_posix()
PDB_cm = (_data_ref / "cryst_then_model.pdb").as_posix()
PDB_cm_gz = (_data_ref / "cryst_then_model.pdb.gz").as_posix()
PDB_cm_bz2 = (_data_ref / "cryst_then_model.pdb.bz2").as_posix()
PDB_mc = (_data_ref / "model_then_cryst.pdb").as_posix()
PDB_mc_gz = (_data_ref / "model_then_cryst.pdb.gz").as_posix()
PDB_mc_bz2 = (_data_ref / "model_then_cryst.pdb.bz2").as_posix()
PDB_chainidnewres = (_data_ref / "chainIDnewres.pdb.gz").as_posix()
PDB_sameresid_diffresname = (
    _data_ref / "sameresid_diffresname.pdb"
).as_posix()
PDB_chainidrepeat = (_data_ref / "chainIDrepeat.pdb.gz").as_posix()
PDB_multiframe = (_data_ref / "nmr_neopetrosiamide.pdb").as_posix()
PDB_helix = (_data_ref / "A6PA6_alpha.pdb").as_posix()
PDB_conect = (_data_ref / "conect_parsing.pdb").as_posix()
PDB_conect2TER = (_data_ref / "CONECT2TER.pdb").as_posix()
PDB_singleconect = (_data_ref / "SINGLECONECT.pdb").as_posix()
PDB_icodes = (_data_ref / "1osm.pdb.gz").as_posix()
PDB_CRYOEM_BOX = (_data_ref / "5a7u.pdb").as_posix()
PDB_CHECK_RIGHTHAND_PA = (_data_ref / "6msm.pdb.bz2").as_posix()
FHIAIMS = (_data_ref / "fhiaims.in").as_posix()

GRO = (_data_ref / "adk_oplsaa.gro").as_posix()
GRO_velocity = (_data_ref / "sample_velocity_file.gro").as_posix()
GRO_incomplete_vels = (_data_ref / "grovels.gro").as_posix()
GRO_large = (_data_ref / "bigbox.gro.bz2").as_posix()
GRO_residwrap = (_data_ref / "residwrap.gro").as_posix()
GRO_residwrap_0base = (_data_ref / "residwrap_0base.gro").as_posix()
GRO_sameresid_diffresname = (
    _data_ref / "sameresid_diffresname.gro"
).as_posix()
PDB = (_data_ref / "adk_oplsaa.pdb").as_posix()
XTC = (_data_ref / "adk_oplsaa.xtc").as_posix()
TRR = (_data_ref / "adk_oplsaa.trr").as_posix()
TPR = (_data_ref / "adk_oplsaa.tpr").as_posix()
PDB_sub_dry = (_data_ref / "cobrotoxin_dry_neutral_0.pdb").as_posix()
TRR_sub_sol = (_data_ref / "cobrotoxin.trr").as_posix()
XTC_sub_sol = (_data_ref / "cobrotoxin.xtc").as_posix()
PDB_sub_sol = (_data_ref / "cobrotoxin.pdb").as_posix()
PDB_xlserial = (_data_ref / "xl_serial.pdb").as_posix()
GRO_MEMPROT = (_data_ref / "analysis/YiiP_lipids.gro.gz").as_posix()
XTC_MEMPROT = (_data_ref / "analysis/YiiP_lipids.xtc").as_posix()
XTC_multi_frame = (
    _data_ref / "xtc_test_only_10_frame_10_atoms.xtc"
).as_posix()
TRR_multi_frame = (
    _data_ref / "trr_test_only_10_frame_10_atoms.trr"
).as_posix()
TNG_traj = (_data_ref / "argon_npt_compressed.tng").as_posix()
TNG_traj_gro = (_data_ref / "argon_npt_compressed.gro.gz").as_posix()
TNG_traj_uneven_blocks = (
    _data_ref / "argon_npt_compressed_uneven.tng"
).as_posix()
TNG_traj_vels_forces = (
    _data_ref / "argon_npt_compressed_vels_forces.tng"
).as_posix()
PDB_xvf = (_data_ref / "cobrotoxin.pdb").as_posix()
TPR_xvf = (_data_ref / "cobrotoxin.tpr").as_posix()
TPR_xvf_2024_4 = (_data_ref / "cobrotoxin_2024_4.tpr").as_posix()
TRR_xvf = (_data_ref / "cobrotoxin.trr").as_posix()
H5MD_xvf = (_data_ref / "cobrotoxin.h5md").as_posix()
H5MD_energy = (_data_ref / "cu.h5md").as_posix()
H5MD_malformed = (_data_ref / "cu_malformed.h5md").as_posix()
XVG_BZ2 = (_data_ref / "cobrotoxin_protein_forces.xvg.bz2").as_posix()

XPDB_small = (_data_ref / "5digitResid.pdb").as_posix()
# number is the gromacs version
TPR400 = (_data_ref / "tprs/2lyz_gmx_4.0.tpr").as_posix()
TPR402 = (_data_ref / "tprs/2lyz_gmx_4.0.2.tpr").as_posix()
TPR403 = (_data_ref / "tprs/2lyz_gmx_4.0.3.tpr").as_posix()
TPR404 = (_data_ref / "tprs/2lyz_gmx_4.0.4.tpr").as_posix()
TPR405 = (_data_ref / "tprs/2lyz_gmx_4.0.5.tpr").as_posix()
TPR406 = (_data_ref / "tprs/2lyz_gmx_4.0.6.tpr").as_posix()
TPR407 = (_data_ref / "tprs/2lyz_gmx_4.0.7.tpr").as_posix()
TPR450 = (_data_ref / "tprs/2lyz_gmx_4.5.tpr").as_posix()
TPR451 = (_data_ref / "tprs/2lyz_gmx_4.5.1.tpr").as_posix()
TPR452 = (_data_ref / "tprs/2lyz_gmx_4.5.2.tpr").as_posix()
TPR453 = (_data_ref / "tprs/2lyz_gmx_4.5.3.tpr").as_posix()
TPR454 = (_data_ref / "tprs/2lyz_gmx_4.5.4.tpr").as_posix()
TPR455 = (_data_ref / "tprs/2lyz_gmx_4.5.5.tpr").as_posix()
TPR502 = (_data_ref / "tprs/2lyz_gmx_5.0.2.tpr").as_posix()
TPR504 = (_data_ref / "tprs/2lyz_gmx_5.0.4.tpr").as_posix()
TPR505 = (_data_ref / "tprs/2lyz_gmx_5.0.5.tpr").as_posix()
TPR510 = (_data_ref / "tprs/2lyz_gmx_5.1.tpr").as_posix()
TPR2016 = (_data_ref / "tprs/2lyz_gmx_2016.tpr").as_posix()
TPR2018 = (_data_ref / "tprs/2lyz_gmx_2018.tpr").as_posix()
TPR2019B3 = (_data_ref / "tprs/2lyz_gmx_2019-beta3.tpr").as_posix()
TPR2020B2 = (_data_ref / "tprs/2lyz_gmx_2020-beta2.tpr").as_posix()
TPR2020 = (_data_ref / "tprs/2lyz_gmx_2020.tpr").as_posix()
TPR2021 = (_data_ref / "tprs/2lyz_gmx_2021.tpr").as_posix()
TPR2022RC1 = (_data_ref / "tprs/2lyz_gmx_2022-rc1.tpr").as_posix()
TPR2023 = (_data_ref / "tprs/2lyz_gmx_2023.tpr").as_posix()
TPR2024 = (_data_ref / "tprs/2lyz_gmx_2024.tpr").as_posix()
TPR2024_4 = (_data_ref / "tprs/2lyz_gmx_2024_4.tpr").as_posix()
TPR2025_0 = (_data_ref / "tprs/2lyz_gmx_2025_0.tpr").as_posix()
# double precision
TPR455Double = (_data_ref / "tprs/drew_gmx_4.5.5.double.tpr").as_posix()
TPR460 = (_data_ref / "tprs/ab42_gmx_4.6.tpr").as_posix()
TPR461 = (_data_ref / "tprs/ab42_gmx_4.6.1.tpr").as_posix()
TPR2020Double = (_data_ref / "tprs/2lyz_gmx_2020_double.tpr").as_posix()
TPR2021Double = (_data_ref / "tprs/2lyz_gmx_2021_double.tpr").as_posix()
# all bonded interactions
TPR334_bonded = (_data_ref / "tprs/all_bonded/dummy_3.3.4.tpr").as_posix()
TPR510_bonded = (_data_ref / "tprs/all_bonded/dummy_5.1.tpr").as_posix()
TPR2016_bonded = (_data_ref / "tprs/all_bonded/dummy_2016.tpr").as_posix()
TPR2018_bonded = (_data_ref / "tprs/all_bonded/dummy_2018.tpr").as_posix()
TPR2019B3_bonded = (
    _data_ref / "tprs/all_bonded/dummy_2019-beta3.tpr"
).as_posix()
TPR2020B2_bonded = (
    _data_ref / "tprs/all_bonded/dummy_2020-beta2.tpr"
).as_posix()
TPR2020_bonded = (_data_ref / "tprs/all_bonded/dummy_2020.tpr").as_posix()
TPR2020_double_bonded = (
    _data_ref / "tprs/all_bonded/dummy_2020_double.tpr"
).as_posix()
TPR2021_bonded = (_data_ref / "tprs/all_bonded/dummy_2021.tpr").as_posix()
TPR2021_double_bonded = (
    _data_ref / "tprs/all_bonded/dummy_2021_double.tpr"
).as_posix()
TPR2022RC1_bonded = (
    _data_ref / "tprs/all_bonded/dummy_2022-rc1.tpr"
).as_posix()
TPR2023_bonded = (_data_ref / "tprs/all_bonded/dummy_2023.tpr").as_posix()
TPR2024_bonded = (_data_ref / "tprs/all_bonded/dummy_2024.tpr").as_posix()
TPR2024_4_bonded = (_data_ref / "tprs/all_bonded/dummy_2024_4.tpr").as_posix()
TPR2025_0_bonded = (_data_ref / "tprs/all_bonded/dummy_2025_0.tpr").as_posix()
# all interactions
TPR_EXTRA_2025_0 = (
    _data_ref / "tprs/virtual_sites/extra-interactions-2025_0.tpr"
).as_posix()
TPR_EXTRA_2024_4 = (
    _data_ref / "tprs/virtual_sites/extra-interactions-2024_4.tpr"
).as_posix()
TPR_EXTRA_2024 = (
    _data_ref / "tprs/virtual_sites/extra-interactions-2024.tpr"
).as_posix()
TPR_EXTRA_2023 = (
    _data_ref / "tprs/virtual_sites/extra-interactions-2023.tpr"
).as_posix()
TPR_EXTRA_2022RC1 = (
    _data_ref / "tprs/virtual_sites/extra-interactions-2022-rc1.tpr"
).as_posix()
TPR_EXTRA_2021 = (
    _data_ref / "tprs/virtual_sites/extra-interactions-2021.tpr"
).as_posix()
TPR_EXTRA_2020 = (
    _data_ref / "tprs/virtual_sites/extra-interactions-2020.tpr"
).as_posix()
TPR_EXTRA_2018 = (
    _data_ref / "tprs/virtual_sites/extra-interactions-2018.tpr"
).as_posix()
TPR_EXTRA_2016 = (
    _data_ref / "tprs/virtual_sites/extra-interactions-2016.3.tpr"
).as_posix()
TPR_EXTRA_407 = (
    _data_ref / "tprs/virtual_sites/extra-interactions-4.0.7.tpr"
).as_posix()
# ALA dipeptide with neural network potential and a few other options
TPR_NNPOT_2025_0 = (_data_ref / "tprs/ala_nnpot_gmx_2025_0.tpr").as_posix()

XYZ_psf = (_data_ref / "2r9r-1b.psf").as_posix()
XYZ_bz2 = (_data_ref / "2r9r-1b.xyz.bz2").as_posix()
XYZ = (_data_ref / "2r9r-1b.xyz").as_posix()
XYZ_mini = (_data_ref / "mini.xyz").as_posix()
XYZ_five = (_data_ref / "five.xyz").as_posix()
TXYZ = (_data_ref / "coordinates/test.txyz").as_posix()
ARC = (_data_ref / "coordinates/test.arc").as_posix()
ARC_PBC = (_data_ref / "coordinates/new_hexane.arc").as_posix()

PRM = (_data_ref / "Amber/ache.prmtop").as_posix()
TRJ = (_data_ref / "Amber/ache.mdcrd").as_posix()
INPCRD = (_data_ref / "Amber/test.inpcrd").as_posix()
TRJ_bz2 = (_data_ref / "Amber/ache.mdcrd.bz2").as_posix()
PFncdf_Top = (_data_ref / "Amber/posfor.top").as_posix()
PFncdf_Trj = (_data_ref / "Amber/posfor.ncdf").as_posix()
PRM_chainid_bz2 = (_data_ref / "Amber/ache_chainid.prmtop.bz2").as_posix()

CPPTRAJ_TRAJ_TOP = (_data_ref / "Amber/cpptraj_traj.prmtop").as_posix()
CPPTRAJ_TRAJ = (_data_ref / "Amber/cpptraj_traj.nc").as_posix()

PRMpbc = (_data_ref / "Amber/capped-ala.prmtop").as_posix()
TRJpbc_bz2 = (_data_ref / "Amber/capped-ala.mdcrd.bz2").as_posix()

PRMncdf = (_data_ref / "Amber/bala.prmtop").as_posix()
TRJncdf = (_data_ref / "Amber/bala.trj").as_posix()
NCDF = (_data_ref / "Amber/bala.ncdf").as_posix()

PRM12 = (_data_ref / "Amber/anti.top").as_posix()
TRJ12_bz2 = (_data_ref / "Amber/anti_md1.mdcrd.bz2").as_posix()

PRM7 = (_data_ref / "Amber/tz2.truncoct.parm7.bz2").as_posix()
NCDFtruncoct = (_data_ref / "Amber/tz2.truncoct.nc").as_posix()

PRMcs = (_data_ref / "Amber/chitosan.prmtop").as_posix()

PRMNCRST = (_data_ref / "Amber/ace_mbondi3.parm7").as_posix()

PRM_NCBOX = (_data_ref / "Amber/ace_tip3p.parm7").as_posix()
TRJ_NCBOX = (_data_ref / "Amber/ace_tip3p.nc").as_posix()

PRMNEGATIVE = (_data_ref / "Amber/ace_mbondi3.negative.parm7").as_posix()

PRMErr1 = (_data_ref / "Amber/ace_mbondi3.error1.parm7").as_posix()
PRMErr2 = (_data_ref / "Amber/ace_mbondi3.error2.parm7").as_posix()
PRMErr3 = (_data_ref / "Amber/ace_mbondi3.error3.parm7").as_posix()
PRMErr4 = (_data_ref / "Amber/ace_mbondi3.error4.parm7").as_posix()
PRMErr5 = (_data_ref / "Amber/ache_chainid.error5.prmtop.bz2").as_posix()

PRM_UreyBradley = (_data_ref / "Amber/parmed_fad.prmtop").as_posix()
PRM7_ala2 = (_data_ref / "Amber/parmed_ala2_solv.parm7").as_posix()
RST7_ala2 = (_data_ref / "Amber/parmed_ala2_solv.rst7").as_posix()

PRM19SBOPC = (_data_ref / "Amber/ala.ff19SB.OPC.parm7.bz2").as_posix()

PQR = (_data_ref / "adk_open.pqr").as_posix()
PQR_icodes = (_data_ref / "1A2C.pqr").as_posix()

PDBQT_input = (_data_ref / "pdbqt_inputpdbqt.pdbqt").as_posix()
PDBQT_querypdb = (_data_ref / "pdbqt_querypdb.pdb").as_posix()
PDBQT_tyrosol = (_data_ref / "tyrosol.pdbqt.bz2").as_posix()

FASTA = (_data_ref / "test.fasta").as_posix()
HELANAL_BENDING_MATRIX = (
    _data_ref / "helanal_bending_matrix_AdK_DIMS_H8.dat"
).as_posix()
HELANAL_BENDING_MATRIX_SUBSET = (
    _data_ref / "helanal_bending_matrix_AdK_DIMS_H8_frames10to79.dat"
).as_posix()

PDB_HOLE = (_data_ref / "1grm_single.pdb").as_posix()
MULTIPDB_HOLE = (_data_ref / "1grm_elNemo_mode7.pdb.bz2").as_posix()

DMS = (_data_ref / "adk_closed.dms").as_posix()
DMS_DOMAINS = (_data_ref / "adk_closed_domains.dms").as_posix()
DMS_NO_SEGID = (_data_ref / "adk_closed_no_segid.dms").as_posix()

CONECT = (_data_ref / "1hvr.pdb").as_posix()
CONECT_ERROR = (_data_ref / "conect_error.pdb").as_posix()

TRZ = (_data_ref / "trzfile.trz").as_posix()
TRZ_psf = (_data_ref / "trz_psf.psf").as_posix()

TRIC = (_data_ref / "dppc_vesicle_hg.gro").as_posix()

PDB_full = (_data_ref / "4E43.pdb").as_posix()

merge_protein = (_data_ref / "merge/2zmm/protein.pdb").as_posix()
merge_ligand = (_data_ref / "merge/2zmm/ligand.pdb").as_posix()
merge_water = (_data_ref / "merge/2zmm/water.pdb").as_posix()

mol2_molecules = (_data_ref / "mol2/Molecules.mol2").as_posix()
mol2_molecule = (_data_ref / "mol2/Molecule.mol2").as_posix()
mol2_ligand = (_data_ref / "mol2/Ligand.mol2").as_posix()
mol2_broken_molecule = (_data_ref / "mol2/BrokenMolecule.mol2").as_posix()
mol2_comments_header = (
    _data_ref / "mol2/Molecule_comments_header.mol2"
).as_posix()
# MOL2 file without substructure field
mol2_zinc = (_data_ref / "mol2/zinc_856218.mol2").as_posix()
# MOL2 file without bonds
mol2_sodium_ion = (_data_ref / "mol2/sodium_ion.mol2").as_posix()

capping_input = (_data_ref / "capping/aaqaa.gro").as_posix()
capping_output = (_data_ref / "capping/maestro_aaqaa_capped.pdb").as_posix()
capping_ace = (_data_ref / "capping/ace.pdb").as_posix()
capping_nma = (_data_ref / "capping/nma.pdb").as_posix()

contacts_villin_folded = (
    _data_ref / "contacts/villin_folded.gro.bz2"
).as_posix()
contacts_villin_unfolded = (
    _data_ref / "contacts/villin_unfolded.gro.bz2"
).as_posix()
contacts_file = (_data_ref / "contacts/2F4K_qlist5_remap.dat").as_posix()

trz4data = (_data_ref / "lammps/datatest.trz").as_posix()
LAMMPSdata = (_data_ref / "lammps/datatest.data").as_posix()
LAMMPSdata_mini = (_data_ref / "lammps/mini.data").as_posix()
LAMMPSdata2 = (_data_ref / "lammps/ifabp_apo_100mM.data.bz2").as_posix()
LAMMPSdcd2 = (_data_ref / "lammps/ifabp_apo_100mM.dcd").as_posix()
LAMMPScnt = (_data_ref / "lammps/cnt-hexagonal-class1.data").as_posix()
LAMMPScnt2 = (_data_ref / "lammps/cnt-hexagonal-class1.data2").as_posix()
LAMMPShyd = (_data_ref / "lammps/hydrogen-class1.data").as_posix()
LAMMPShyd2 = (_data_ref / "lammps/hydrogen-class1.data2").as_posix()
LAMMPSdata_deletedatoms = (_data_ref / "lammps/deletedatoms.data").as_posix()
LAMMPSdata_triclinic = (_data_ref / "lammps/albite_triclinic.data").as_posix()
LAMMPSdata_PairIJ = (_data_ref / "lammps/pairij_coeffs.data.bz2").as_posix()
LAMMPSDUMP = (_data_ref / "lammps/wat.lammpstrj.bz2").as_posix()
LAMMPSDUMP_long = (_data_ref / "lammps/wat.lammpstrj_long.bz2").as_posix()
LAMMPSDUMP_allinfo = (_data_ref / "lammps/mass_q_elem.lammpstrj").as_posix()
LAMMPSDUMP_nomass_elemx = (
    _data_ref / "lammps/nomass_elemx.lammpstrj"
).as_posix()
LAMMPSDUMP_allcoords = (
    _data_ref / "lammps/spce_all_coords.lammpstrj.bz2"
).as_posix()
LAMMPSDUMP_nocoords = (
    _data_ref / "lammps/spce_no_coords.lammpstrj.bz2"
).as_posix()
LAMMPSDUMP_triclinic = (_data_ref / "lammps/albite_triclinic.dump").as_posix()
LAMMPSDUMP_image_vf = (_data_ref / "lammps/image_vf.lammpstrj").as_posix()
LAMMPS_image_vf = (_data_ref / "lammps/image_vf.data").as_posix()
LAMMPSDUMP_chain1 = (_data_ref / "lammps/chain_dump_1.lammpstrj").as_posix()
LAMMPSDUMP_chain2 = (_data_ref / "lammps/chain_dump_2.lammpstrj").as_posix()
LAMMPS_chain = (_data_ref / "lammps/chain_initial.data").as_posix()
LAMMPSdata_many_bonds = (
    _data_ref / "lammps/a_lot_of_bond_types.data"
).as_posix()
LAMMPSdata_additional_columns = (
    _data_ref / "lammps/additional_columns.data"
).as_posix()
LAMMPSDUMP_additional_columns = (
    _data_ref / "lammps/additional_columns.lammpstrj"
).as_posix()
LAMMPSDUMP_non_linear = (
    _data_ref / "analysis/msd/test_non_linear.dump.bz2"
).as_posix()

unordered_res = (_data_ref / "unordered_res.pdb").as_posix()

GMS_ASYMOPT = (_data_ref / "gms/c1opt.gms.gz").as_posix()
GMS_SYMOPT = (_data_ref / "gms/symopt.gms").as_posix()
GMS_ASYMSURF = (_data_ref / "gms/surf2wat.gms").as_posix()

two_water_gro = (_data_ref / "two_water_gro.gro").as_posix()
two_water_gro_multiframe = (
    _data_ref / "two_water_gro_multiframe.gro"
).as_posix()
two_water_gro_nonames = (_data_ref / "two_water_gro_nonames.gro").as_posix()
two_water_gro_widebox = (_data_ref / "two_water_gro_widebox.gro").as_posix()

DLP_CONFIG = (_data_ref / "dlpoly/CONFIG").as_posix()
DLP_CONFIG_order = (_data_ref / "dlpoly/CONFIG_order").as_posix()
DLP_CONFIG_minimal = (_data_ref / "dlpoly/CONFIG_minimal").as_posix()
DLP_HISTORY = (_data_ref / "dlpoly/HISTORY").as_posix()
DLP_HISTORY_order = (_data_ref / "dlpoly/HISTORY_order").as_posix()
DLP_HISTORY_minimal = (_data_ref / "dlpoly/HISTORY_minimal").as_posix()
DLP_HISTORY_minimal_cell = (
    _data_ref / "dlpoly/HISTORY_minimal_cell"
).as_posix()
DLP_HISTORY_classic = (_data_ref / "dlpoly/HISTORY_classic").as_posix()

waterPSF = (_data_ref / "watdyn.psf").as_posix()
waterDCD = (_data_ref / "watdyn.dcd").as_posix()

rmsfArray = (_data_ref / "adk_oplsaa_CA_rmsf.npy").as_posix()

HoomdXMLdata = (_data_ref / "C12x64.xml.bz2").as_posix()

Make_Whole = (_data_ref / "make_whole.gro").as_posix()
fullerene = (_data_ref / "fullerene.pdb.gz").as_posix()

Plength = (_data_ref / "plength.gro").as_posix()
Martini_membrane_gro = (_data_ref / "martini_dppc_chol_bilayer.gro").as_posix()

# Contains one of each residue in 'nucleic' selections
NUCLsel = (_data_ref / "nucl_res.pdb").as_posix()

RANDOM_WALK = (_data_ref / "xyz_random_walk.xtc").as_posix()
RANDOM_WALK_TOPO = (_data_ref / "RANDOM_WALK_TOPO.pdb").as_posix()

MMTF = (_data_ref / "173D.mmtf").as_posix()
MMTF_gz = (_data_ref / "5KIH.mmtf.gz").as_posix()
MMTF_skinny = (_data_ref / "1ubq-less-optional.mmtf").as_posix()
MMTF_skinny2 = (_data_ref / "3NJW-onlyrequired.mmtf").as_posix()
MMTF_NOCRYST = (_data_ref / "6QYR.mmtf.gz").as_posix()

ALIGN_BOUND = (_data_ref / "analysis/align_bound.pdb.gz").as_posix()
ALIGN_UNBOUND = (_data_ref / "analysis/align_unbound.pdb.gz").as_posix()

GSD = (_data_ref / "example.gsd").as_posix()
GSD_bonds = (_data_ref / "example_bonds.gsd").as_posix()
GSD_long = (_data_ref / "example_longer.gsd").as_posix()

TRC_PDB_VAC = (_data_ref / "gromos11/gromos11_traj_vac.pdb.gz").as_posix()
TRC_TRAJ1_VAC = (_data_ref / "gromos11/gromos11_traj_vac_1.trc.gz").as_posix()
TRC_TRAJ1_VAC_WHITESPACE = (
    _data_ref / "gromos11/gromos11_traj_vac_1_whitespace.trc.gz"
).as_posix()
TRC_TRAJ1_VAC_MISSING_POS = (
    _data_ref / "gromos11/gromos11_traj_vac_1_missing_pos.trc.gz"
).as_posix()
TRC_TRAJ1_VAC_EXTRA_POS = (
    _data_ref / "gromos11/gromos11_traj_vac_1_extra_pos.trc.gz"
).as_posix()
TRC_TRAJ2_VAC = (_data_ref / "gromos11/gromos11_traj_vac_2.trc.gz").as_posix()
TRC_PDB_SOLV = (_data_ref / "gromos11/gromos11_traj_solv.pdb.gz").as_posix()
TRC_TRAJ_SOLV = (_data_ref / "gromos11/gromos11_traj_solv.trc.gz").as_posix()
TRC_CLUSTER_VAC = (
    _data_ref / "gromos11/gromos11_cluster_vac.trj.gz"
).as_posix()
TRC_TRICLINIC_SOLV = (
    _data_ref / "gromos11/gromos11_triclinic_solv.trc.gz"
).as_posix()
TRC_TRUNCOCT_VAC = (
    _data_ref / "gromos11/gromos11_truncOcta_vac.trc.gz"
).as_posix()
TRC_GENBOX_ORIGIN = (
    _data_ref / "gromos11/gromos11_genbox_origin.trc.gz"
).as_posix()
TRC_GENBOX_EULER = (
    _data_ref / "gromos11/gromos11_genbox_euler.trc.gz"
).as_posix()
TRC_EMPTY = (_data_ref / "gromos11/gromos11_empty.trc").as_posix()

DihedralArray = (_data_ref / "adk_oplsaa_dihedral.npy").as_posix()
DihedralsArray = (_data_ref / "adk_oplsaa_dihedral_list.npy").as_posix()
RamaArray = (_data_ref / "adk_oplsaa_rama.npy").as_posix()
GLYRamaArray = (_data_ref / "adk_oplsaa_GLY_rama.npy").as_posix()
JaninArray = (_data_ref / "adk_oplsaa_janin.npy").as_posix()
LYSJaninArray = (_data_ref / "adk_oplsaa_LYS_janin.npy").as_posix()
PDB_rama = (_data_ref / "19hc.pdb.gz").as_posix()
PDB_janin = (_data_ref / "1a28.pdb.gz").as_posix()

BATArray = (_data_ref / "mol2_comments_header_bat.npy").as_posix()

ITP = (_data_ref / "gromacs_ala10.itp").as_posix()
ITP_nomass = (_data_ref / "itp_nomass.itp").as_posix()
ITP_atomtypes = (_data_ref / "atomtypes.itp").as_posix()
ITP_charges = (_data_ref / "atomtypes_charge.itp").as_posix()
ITP_edited = (_data_ref / "edited_itp.itp").as_posix()
ITP_tip5p = (_data_ref / "tip5p.itp").as_posix()
ITP_spce = (_data_ref / "spce.itp").as_posix()

GMX_TOP = (_data_ref / "gromacs_ala10.top").as_posix()
GMX_DIR = (_data_ref / "gromacs/").as_posix()
GMX_TOP_BAD = (_data_ref / "bad_top.top").as_posix()
ITP_no_endif = (_data_ref / "no_endif_spc.itp").as_posix()

NAMDBIN = (_data_ref / "adk_open.coor").as_posix()

SDF_molecule = (_data_ref / "molecule.sdf").as_posix()

PDB_elements = (_data_ref / "elements.pdb").as_posix()
PDB_charges = (_data_ref / "charges.pdb").as_posix()

PDBX = (_data_ref / "4x8u.pdbx").as_posix()

SURFACE_PDB = (_data_ref / "surface.pdb.bz2").as_posix()
SURFACE_TRR = (_data_ref / "surface.trr").as_posix()

# DSSP testing: from https://github.com/ShintaroMinami/PyDSSP
DSSP = (_data_ref / "dssp").as_posix()

# This should be the last line: clean up namespace
del resources



# -*- Mode: python; tab-width: 4; indent-tabs-mode:nil; coding:utf-8 -*-
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4
#
# MDAnalysis --- https://www.mdanalysis.org
# Copyright (c) 2006-2017 The MDAnalysis Development Team and contributors
# (see the file AUTHORS for the full list of names)
#
# Released under the Lesser GNU Public Licence, v2.1 or any higher version
#
# Please cite your use of MDAnalysis in published work:
#
# R. J. Gowers, M. Linke, J. Barnoud, T. J. E. Reddy, M. N. Melo, S. L. Seyler,
# D. L. Dotson, J. Domanski, S. Buchoux, I. M. Kenney, and O. Beckstein.
# MDAnalysis: A Python package for the rapid analysis of molecular dynamics
# simulations. In S. Benthall and S. Rostrup editors, Proceedings of the 15th
# Python in Science Conference, pages 102-109, Austin, TX, 2016. SciPy.
# doi: 10.25080/majora-629e541a-00e
#
# N. Michaud-Agrawal, E. J. Denning, T. B. Woolf, and O. Beckstein.
# MDAnalysis: A Toolkit for the Analysis of Molecular Dynamics Simulations.
# J. Comput. Chem. 32 (2011), 2319--2327, doi:10.1002/jcc.21787
#
"""GSD trajectory reader  --- :mod:`MDAnalysis.coordinates.GSD`
============================================================

Class to read the GSD trajectory, output of `HOOMD-blue`_. The GSD format
specifies both the topology and the trajectory of the particles in the
simulation. The topology is read by the
:class:`~MDAnalysis.topology.GSDParser.GSDParser` class.

The GSD format was developed having in mind the possibility of changing number
of particles, particle types, particle identities and changing topology.
Currently this class has limited functionality, due to the fact that the number
of particles and the topology are kept fixed in most MD simulations. The user
will get an error only if at any time step the number of particles is detected
to be different to the one that was set at the first time step. No check on
changes in particle identity or topology is currently implemented.

.. _`HOOMD-blue`: https://glotzerlab.engin.umich.edu/hoomd-blue/

Classes
-------

.. autoclass:: GSDReader
   :inherited-members:

.. autoclass:: GSDPicklable
   :members:

.. autofunction:: gsd_pickle_open

"""
import numpy as np

try:
    import gsd
    import gsd.fl
except ImportError:
    HAS_GSD = False
    import types

    class MockHOOMDTrajectory:
        pass

    gsd = types.ModuleType("gsd")
    gsd.hoomd = types.ModuleType("hoomd")
    gsd.hoomd.HOOMDTrajectory = MockHOOMDTrajectory
else:
    HAS_GSD = True

from . import base
from MDAnalysis.lib.util import store_init_arguments


class GSDReader(base.ReaderBase):
    """Reader for the GSD format."""

    format = "GSD"
    units = {"time": None, "length": None}

    @store_init_arguments
    def __init__(self, filename, **kwargs):
        """
        Parameters
        ----------
        filename : str
            trajectory filename
        **kwargs : dict
            General reader arguments.


        .. versionadded:: 0.17.0
        .. versionchanged:: 2.0.0
            Now use a picklable :class:`gsd.hoomd.HOOMDTrajectory`--
            :class:`GSDPicklable`
        .. versionchanged:: 2.6.0
           Support for GSD versions below 3.0.1 have been dropped. This
           includes support for schema 1.3.
        """
        if not HAS_GSD:
            errmsg = "GSDReader: To read GSD files, please install gsd"
            raise ImportError(errmsg)

        super(GSDReader, self).__init__(filename, **kwargs)
        self.filename = filename
        self.open_trajectory()
        self.n_atoms = self._file[0].particles.N
        self.ts = self._Timestep(self.n_atoms, **self._ts_kwargs)
        self._read_next_timestep()

    def open_trajectory(self):
        """opens the trajectory file using gsd.hoomd module"""
        self._frame = -1
        self._file = gsd_pickle_open(self.filename, mode="r")

    def close(self):
        """close reader"""
        self._file.file.close()

    @property
    def n_frames(self):
        """number of frames in trajectory"""
        return len(self._file)

    def _reopen(self):
        """reopen trajectory"""
        self.close()
        self.open_trajectory()

    def _read_frame(self, frame):
        try:
            myframe = self._file[frame]
        except IndexError:
            raise IOError from None

        # set frame number
        self._frame = frame

        # sets the Timestep object
        self.ts.frame = frame
        self.ts.data["step"] = myframe.configuration.step

        # set frame box dimensions
        self.ts.dimensions = myframe.configuration.box
        self.ts.dimensions[3:] = np.rad2deg(np.arccos(self.ts.dimensions[3:]))

        # set particle positions
        frame_positions = myframe.particles.position
        n_atoms_now = frame_positions.shape[0]
        if n_atoms_now != self.n_atoms:
            raise ValueError(
                "Frame %d has %d atoms but the initial frame has %d"
                " atoms. MDAnalysis in unable to deal with variable"
                " topology!" % (frame, n_atoms_now, self.n_atoms)
            )
        else:
            self.ts.positions = frame_positions
        return self.ts

    def _read_next_timestep(self):
        """read next frame in trajectory"""
        return self._read_frame(self._frame + 1)


class GSDPicklable(gsd.hoomd.HOOMDTrajectory):
    """Hoomd GSD file object (read-only) that can be pickled.

    This class provides a file-like object (as by :func:`gsd.hoomd.open`,
    namely :class:`gsd.hoodm.HOOMDTrajectory`) that, unlike file objects,
    can be pickled. Only read mode is supported.

    When the file is pickled, filename and mode of :class:`gsd.fl.GSDFile` in
    the file are saved. On unpickling, the file is opened by filename.
    This means that for a successful unpickle, the original file still has to
    be accessible with its filename.

    Note
    ----
    Open hoomd GSD files with `gsd_pickle_open`.
    After pickling, the current frame is reset. `universe.trajectory[i]` has
    to be used to return to its original frame.

    Parameters
    ----------
    file: :class:`gsd.fl.GSDFile`
        File to access.

    Example
    -------
    ::

        gsdfileobj = gsd.fl.open(name=filename,
                                     mode='r',
                                     application='gsd.hoomd '+ gsd.version.version,
                                     schema='hoomd',
                                     schema_version=[1, 3])
        file = GSDPicklable(gsdfileobj)
        file_pickled = pickle.loads(pickle.dumps(file))

    See Also
    ---------
    :func:`MDAnalysis.lib.picklable_file_io.FileIOPicklable`
    :func:`MDAnalysis.lib.picklable_file_io.BufferIOPicklable`
    :func:`MDAnalysis.lib.picklable_file_io.TextIOPicklable`
    :func:`MDAnalysis.lib.picklable_file_io.GzipPicklable`
    :func:`MDAnalysis.lib.picklable_file_io.BZ2Picklable`


    .. versionadded:: 2.0.0
    """

    def __getstate__(self):
        return self.file.name, self.file.mode

    def __setstate__(self, args):
        gsd_version = gsd.version.version
        schema_version = [1, 4]
        gsdfileobj = gsd.fl.open(
            name=args[0],
            mode=args[1],
            application="gsd.hoomd " + gsd_version,
            schema="hoomd",
            schema_version=schema_version,
        )
        self.__init__(gsdfileobj)


def gsd_pickle_open(name: str, mode: str = "r"):
    """Open hoomd schema GSD file with pickle function implemented.

    This function returns a GSDPicklable object. It can be used as a
    context manager, and replace the built-in :func:`gsd.hoomd.open` function
    in read mode that only returns an unpicklable file object.

    Schema version will depend on the version of gsd module.

    Note
    ----
    Can be only used with read mode.

    Parameters
    ----------
    name : str
        a filename given a text or byte string.
    mode: str, optional
        'r':  open for reading.

    Returns
    -------
    stream-like object: GSDPicklable

    Raises
    ------
    ValueError
        if `mode` is not one of the allowed read modes

    Examples
    -------
    open as context manager::

        with gsd_pickle_open('filename') as f:
            line = f.readline()

    open as function::

        f = gsd_pickle_open('filename')
        line = f.readline()
        f.close()

    See Also
    --------
    :func:`MDAnalysis.lib.util.anyopen`
    :func:`MDAnalysis.lib.picklable_file_io.pickle_open`
    :func:`MDAnalysis.lib.picklable_file_io.bz2_pickle_open`
    :func:`MDAnalysis.lib.picklable_file_io.gzip_pickle_open`
    :func:`gsd.hoomd.open`


    .. versionadded:: 2.0.0
    .. versionchanged:: 2.6.0
       Only GSD versions 3.0.1+ are supported. 'rb' mode support
       has been replaced with 'r' mode.
    """
    gsd_version = gsd.version.version
    schema_version = [1, 4]
    if mode != "r":
        raise ValueError("Only read mode 'r' " "files can be pickled.")
    gsdfileobj = gsd.fl.open(
        name=name,
        mode=mode,
        application="gsd.hoomd " + gsd_version,
        schema="hoomd",
        schema_version=schema_version,
    )
    return GSDPicklable(gsdfileobj)



# -*- Mode: python; tab-width: 4; indent-tabs-mode:nil; coding:utf-8 -*-
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4 fileencoding=utf-8
#
# MDAnalysis --- https://www.mdanalysis.org
# Copyright (c) 2006-2017 The MDAnalysis Development Team and contributors
# (see the file AUTHORS for the full list of names)
#
# Released under the Lesser GNU Public Licence, v2.1 or any higher version
#
# Please cite your use of MDAnalysis in published work:
#
# R. J. Gowers, M. Linke, J. Barnoud, T. J. E. Reddy, M. N. Melo, S. L. Seyler,
# D. L. Dotson, J. Domanski, S. Buchoux, I. M. Kenney, and O. Beckstein.
# MDAnalysis: A Python package for the rapid analysis of molecular dynamics
# simulations. In S. Benthall and S. Rostrup editors, Proceedings of the 15th
# Python in Science Conference, pages 102-109, Austin, TX, 2016. SciPy.
# doi: 10.25080/majora-629e541a-00e
#
# N. Michaud-Agrawal, E. J. Denning, T. B. Woolf, and O. Beckstein.
# MDAnalysis: A Toolkit for the Analysis of Molecular Dynamics Simulations.
# J. Comput. Chem. 32 (2011), 2319--2327, doi:10.1002/jcc.21787
#
import pytest
import numpy as np

import MDAnalysis as mda

from MDAnalysisTests.topology.base import ParserBase
from MDAnalysisTests.datafiles import CONECT, PDBX, PDB


try:
    from openmm.app import Element, Topology
    from openmm.unit import daltons
    from openmm import app
except ImportError:
    try:
        from simtk.openmm.app import Element, Topology
        from simtk.unit import daltons
        from simtk.openmm import app
    except ImportError:
        pytest.skip(allow_module_level=True)


class OpenMMTopologyBase(ParserBase):
    parser = mda.converters.OpenMMParser.OpenMMTopologyParser
    expected_attrs = [
        "ids",
        "names",
        "resids",
        "resnames",
        "masses",
        "bonds",
        "chainIDs",
        "elements",
        "types",
    ]
    expected_n_bonds = 0

    @pytest.fixture()
    def top(self, filename):
        with self.parser(filename) as p:
            yield p.parse()

    def test_creates_universe(self, filename):
        """Check that Universe works with this Parser"""
        u = mda.Universe(filename, topology_format="OPENMMTOPOLOGY")
        assert isinstance(u, mda.Universe)

    def test_attr_size(self, top):
        assert len(top.ids) == top.n_atoms
        assert len(top.names) == top.n_atoms
        assert len(top.resids) == top.n_residues
        assert len(top.resnames) == top.n_residues

    def test_atoms(self, top):
        assert top.n_atoms == self.expected_n_atoms

    def test_bonds(self, top):
        assert len(top.bonds.values) == self.expected_n_bonds
        if self.expected_n_bonds:
            assert isinstance(top.bonds.values[0], tuple)
        else:
            assert top.bonds.values == []

    def test_resids(self, top):
        assert len(top.resids.values) == self.expected_n_residues
        if self.expected_n_residues:
            assert isinstance(top.resids.values, np.ndarray)
        else:
            assert top.resids.values == []

    def test_resnames(self, top):
        assert len(top.resnames.values) == self.expected_n_residues
        if self.expected_n_residues:
            assert isinstance(top.resnames.values, np.ndarray)
        else:
            assert top.resnames.values == []

    def test_resnums(self, top):
        assert len(top.resnums.values) == self.expected_n_residues
        if self.expected_n_residues:
            assert isinstance(top.resnums.values, np.ndarray)
        else:
            assert top.resnums.values == []

    def test_segids(self, top):
        assert len(top.segids.values) == self.expected_n_segments
        assert all(isinstance(segid, str) for segid in top.segids.values)
        if self.expected_n_segments:
            assert isinstance(top.segids.values, np.ndarray)
        else:
            assert top.segids.values == []

    def test_elements(self, top):
        if "elements" in self.expected_attrs:
            assert len(top.elements.values) == self.expected_n_atoms
            assert isinstance(top.elements.values, np.ndarray)
            assert all(isinstance(elem, str) for elem in top.elements.values)
        else:
            assert not hasattr(top, "elements")

    def test_atomtypes(self, top):
        assert len(top.types.values) == self.expected_n_atoms
        if self.expected_n_atoms:
            assert isinstance(top.types.values, np.ndarray)
        else:
            assert top.types.values == []

    def test_masses(self, top):
        assert len(top.masses.values) == self.expected_n_atoms
        if self.expected_n_atoms:
            assert isinstance(top.masses.values, np.ndarray)
            assert all(
                isinstance(mass, np.float64) for mass in top.masses.values
            )
        else:
            assert top.masses.values == []

    def test_guessed_attributes(self, filename):
        u = mda.Universe(filename, topology_format="OPENMMTOPOLOGY")
        u_guessed_attrs = [
            attr.attrname for attr in u._topology.guessed_attributes
        ]
        for attr in self.guessed_attrs:
            assert hasattr(u.atoms, attr)
            assert attr in u_guessed_attrs


class OpenMMAppTopologyBase(OpenMMTopologyBase):
    parser = mda.converters.OpenMMParser.OpenMMAppTopologyParser
    expected_attrs = [
        "ids",
        "names",
        "resids",
        "resnames",
        "masses",
        "bonds",
        "chainIDs",
        "elements",
        "types",
    ]
    expected_n_bonds = 0

    @pytest.fixture()
    def top(self, filename):
        with self.parser(filename) as p:
            yield p.parse()

    def test_creates_universe(self, filename):
        """Check that Universe works with this Parser"""
        u = mda.Universe(filename, topology_format="OPENMMAPP")
        assert isinstance(u, mda.Universe)

    def test_guessed_attributes(self, filename):
        u = mda.Universe(filename, topology_format="OPENMMAPP")
        for attr in self.guessed_attrs:
            assert hasattr(u.atoms, attr)


class TestOpenMMTopologyParser(OpenMMTopologyBase):
    ref_filename = app.PDBFile(CONECT).topology
    expected_n_atoms = 1890
    expected_n_residues = 199
    expected_n_segments = 3
    expected_n_bonds = 1922


class TestOpenMMTopologyParserWithPartialElements(OpenMMTopologyBase):
    parser = mda.converters.OpenMMParser.OpenMMTopologyParser
    ref_filename = app.PDBFile(PDB).topology
    expected_n_atoms = 47681
    expected_n_residues = 11302
    expected_n_segments = 1
    expected_n_bonds = 25533

    def test_with_partial_elements(self):
        wmsg1 = (
            "Element information missing for some atoms. "
            "These have been given an empty element record "
        )

        wmsg2 = (
            "For absent elements, atomtype has been  "
            "set to 'X' and mass has been set to 0.0. "
            "If needed these can be guessed using "
            "universe.guess_TopologyAttrs(to_guess=['masses', 'types']). "
            "(for MDAnalysis version 2.x this is done automatically,"
            " but it will be removed in 3.0)."
        )

        with pytest.warns(UserWarning) as warnings:
            mda_top = self.parser(self.ref_filename).parse()
            assert mda_top.types.values[3344] == "X"
            assert mda_top.types.values[3388] == "X"
            assert mda_top.elements.values[3344] == ""
            assert mda_top.elements.values[3388] == ""
            assert mda_top.masses.values[3344] == 0.0
            assert mda_top.masses.values[3388] == 0.0

            assert len(warnings) == 2
            assert str(warnings[0].message) == wmsg1
            assert str(warnings[1].message) == wmsg2


def test_no_elements_warn():
    parser = mda.converters.OpenMMParser.OpenMMTopologyParser
    omm_top = app.PDBFile(CONECT).topology
    for a in omm_top.atoms():
        a.element = None

    wmsg = (
        "Element information is missing for all the atoms. "
        "Elements attribute will not be populated. "
        "Atomtype attribute will be guessed using atom "
        "name and mass will be guessed using atomtype."
        "For MDAnalysis version 2.x this is done automatically, "
        "but it will be removed in MDAnalysis v3.0. "
        "These can be guessed using "
        "universe.guess_TopologyAttrs(to_guess=['masses', 'types']) "
        "See MDAnalysis.guessers."
    )

    with pytest.warns(UserWarning) as warnings:
        mda_top = parser(omm_top).parse()
        assert str(warnings[0].message) == wmsg


def test_invalid_element_symbols():
    parser = mda.converters.OpenMMParser.OpenMMTopologyParser
    omm_top = Topology()
    bad1 = Element(0, "*", "*", 0 * daltons)
    bad2 = Element(0, "?", "?", 6 * daltons)
    bad3 = None
    silver = Element.getBySymbol("Ag")
    chain = omm_top.addChain()
    res = omm_top.addResidue("R", chain)
    omm_top.addAtom(name="Ag", element=silver, residue=res)
    omm_top.addAtom(name="Bad1", element=bad1, residue=res)
    omm_top.addAtom(name="Bad2", element=bad2, residue=res)
    omm_top.addAtom(name="Bad3", element=bad3, residue=res)
    mda_top = parser(omm_top).parse()

    assert mda_top.types.values[0] == "Ag"
    assert mda_top.types.values[1] == "*"
    assert mda_top.types.values[2] == "?"
    assert mda_top.types.values[3] == "X"
    assert mda_top.elements.values[0] == "Ag"
    assert mda_top.elements.values[1] == ""
    assert mda_top.elements.values[2] == ""
    assert mda_top.elements.values[3] == ""
    assert mda_top.masses.values[0] == 107.86822
    assert mda_top.masses.values[1] == 0.0
    assert mda_top.masses.values[2] == 6.0
    assert mda_top.masses.values[3] == 0.0


class TestOpenMMPDBFileParser(OpenMMAppTopologyBase):
    ref_filename = app.PDBFile(CONECT)
    expected_n_atoms = 1890
    expected_n_residues = 199
    expected_n_segments = 3
    expected_n_bonds = 1922


class TestOpenMMPDBxFileParser(OpenMMAppTopologyBase):
    ref_filename = app.PDBxFile(PDBX)
    expected_n_atoms = 60
    expected_n_residues = 7
    expected_n_segments = 1
    expected_n_bonds = 62



# -*- Mode: python; tab-width: 4; indent-tabs-mode:nil; coding:utf-8 -*-
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4 fileencoding=utf-8
#
# MDAnalysis --- https://www.mdanalysis.org
# Copyright (c) 2006-2017 The MDAnalysis Development Team and contributors
# (see the file AUTHORS for the full list of names)
#
# Released under the Lesser GNU Public Licence, v2.1 or any higher version
#
# Please cite your use of MDAnalysis in published work:
#
# R. J. Gowers, M. Linke, J. Barnoud, T. J. E. Reddy, M. N. Melo, S. L. Seyler,
# D. L. Dotson, J. Domanski, S. Buchoux, I. M. Kenney, and O. Beckstein.
# MDAnalysis: A Python package for the rapid analysis of molecular dynamics
# simulations. In S. Benthall and S. Rostrup editors, Proceedings of the 15th
# Python in Science Conference, pages 102-109, Austin, TX, 2016. SciPy.
# doi: 10.25080/majora-629e541a-00e
#
# N. Michaud-Agrawal, E. J. Denning, T. B. Woolf, and O. Beckstein.
# MDAnalysis: A Toolkit for the Analysis of Molecular Dynamics Simulations.
# J. Comput. Chem. 32 (2011), 2319--2327, doi:10.1002/jcc.21787
#
import os

import pytest
from numpy.testing import assert_almost_equal

import MDAnalysis as mda
from MDAnalysis.coordinates.GSD import GSDReader, HAS_GSD
from MDAnalysisTests.datafiles import GSD


@pytest.mark.skipif(HAS_GSD, reason="gsd is installed")
def test_no_gsd_u_raises():
    with pytest.raises(ImportError, match="please install gsd"):
        _ = mda.Universe(GSD)


@pytest.mark.skipif(HAS_GSD, reason="gsd is installed")
def test_gsd_reader_raises():
    with pytest.raises(ImportError, match="please install gsd"):
        _ = GSDReader("foo")


@pytest.mark.skipif(not HAS_GSD, reason="gsd is not installed")
class TestGSDReader:
    @pytest.fixture(scope="class")
    def GSD_U(self):
        return mda.Universe(GSD)

    def test_gsd_positions(self, GSD_U):
        # first frame first particle
        GSD_U.trajectory[0]
        assert_almost_equal(
            GSD_U.atoms.positions[0], [-5.4000001, -10.19999981, -10.19999981]
        )
        # second frame first particle
        GSD_U.trajectory[1]
        assert_almost_equal(
            GSD_U.atoms.positions[0], [-5.58348083, -9.98546982, -10.17657185]
        )

    def test_gsd_n_frames(self, GSD_U):
        assert len(GSD_U.trajectory) == 2

    def test_gsd_dimensions(self, GSD_U):
        ts = GSD_U.trajectory[0]
        assert_almost_equal(
            ts.dimensions,
            [21.60000038, 21.60000038, 21.60000038, 90.0, 90.0, 90.0],
        )

    def test_gsd_data_step(self, GSD_U):
        assert GSD_U.trajectory[0].data["step"] == 0
        assert GSD_U.trajectory[1].data["step"] == 500



# -*- Mode: python; tab-width: 4; indent-tabs-mode:nil; coding:utf-8 -*-
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4
#
# MDAnalysis --- https://www.mdanalysis.org
# Copyright (c) 2006-2017 The MDAnalysis Development Team and contributors
# (see the file AUTHORS for the full list of names)
#
# Released under the Lesser GNU Public Licence, v2.1 or any higher version
#
# Please cite your use of MDAnalysis in published work:
#
# R. J. Gowers, M. Linke, J. Barnoud, T. J. E. Reddy, M. N. Melo, S. L. Seyler,
# D. L. Dotson, J. Domanski, S. Buchoux, I. M. Kenney, and O. Beckstein.
# MDAnalysis: A Python package for the rapid analysis of molecular dynamics
# simulations. In S. Benthall and S. Rostrup editors, Proceedings of the 15th
# Python in Science Conference, pages 102-109, Austin, TX, 2016. SciPy.
#
# N. Michaud-Agrawal, E. J. Denning, T. B. Woolf, and O. Beckstein.
# MDAnalysis: A Toolkit for the Analysis of Molecular Dynamics Simulations.
# J. Comput. Chem. 32 (2011), 2319--2327, doi:10.1002/jcc.21787
#

"""\
Transformations Base Class --- :mod:`MDAnalysis.transformations.base`
=====================================================================

.. autoclass:: TransformationBase

"""
from threadpoolctl import threadpool_limits


class TransformationBase(object):
    """Base class for defining on-the-fly transformations

    The class is designed as a template for creating on-the-fly
    Transformation classes. This class will

    1) set up a context manager framework on limiting the threads
    per call, which may be the multi-thread OpenBlas backend of NumPy.
    This backend may kill the performance when subscribing hyperthread
    or oversubscribing the threads when used together with other parallel
    engines e.g. Dask.
    (PR `#2950 <https://github.com/MDAnalysis/mdanalysis/pull/2950>`_)

    Define ``max_threads=1`` when that is the case.

    2) set up a boolean attribute `parallelizable` for checking if the
    transformation can be applied in a **split-apply-combine** parallelism.
    For example, the :class:`~MDAnalysis.transformations.positionaveraging.PositionAverager`
    is history-dependent and can not be used in parallel analysis natively.
    (Issue `#2996 <https://github.com/MDAnalysis/mdanalysis/issues/2996>`_)

    To define a new Transformation, :class:`TransformationBase`
    has to be subclassed.
    ``max_threads`` will be set to ``None`` by default,
    i.e. does not do anything and any settings in the environment such as
    the environment variable :envvar:`OMP_NUM_THREADS`
    (see the `OpenMP specification for OMP_NUM_THREADS <https://www.openmp.org/spec-html/5.0/openmpse50.html>`_)
    are used.
    ``parallelizable`` will be set to ``True`` by default.
    You may need to double check if it can be used in parallel analysis;
    if not, override the value to ``False``.
    Note this attribute is not checked anywhere in MDAnalysis yet.
    Developers of the parallel analysis have to check it in their own code.

    .. code-block:: python

       class NewTransformation(TransformationBase):
           def __init__(self, ag, parameter,
                        max_threads=1, parallelizable=True):
               super().__init__(max_threads=max_threads,
                                 parallelizable=parallelizable)
               self.ag = ag
               self._param = parameter

           def _transform(self, ts):
               #  REQUIRED
               ts.positions = some_function(ts, self.ag, self._param)
               return ts

    Afterwards the new transformation can be run like this.

    .. code-block:: python

       new_transformation = NewTransformation(ag, param)
       u.trajectory.add_transformations(new_transformation)


    .. versionadded:: 2.0.0
       Add the base class for all transformations to limit threads and
       check if it can be used in parallel analysis.
    """

    def __init__(self, **kwargs):
        """
        Parameters
        ----------
        max_threads: int, optional
           The maximum thread number can be used.
           Default is ``None``, which means the default or the external setting.
        parallelizable: bool, optional
           A check for if this can be used in split-apply-combine parallel
           analysis approach.
           Default is ``True``.
        """
        self.max_threads = kwargs.pop("max_threads", None)
        self.parallelizable = kwargs.pop("parallelizable", True)

    def __call__(self, ts):
        """The function that makes transformation can be called as a function

        The thread limit works as a context manager with given `max_threads`
        wrapping the real :func:`_transform` function
        """
        with threadpool_limits(self.max_threads):
            return self._transform(ts)

    def _transform(self, ts):
        """Transform the given `Timestep`

        It deals with the transformation of a single `Timestep`.
        """
        raise NotImplementedError("Only implemented in child classes")



# -*- Mode: python; tab-width: 4; indent-tabs-mode:nil; coding: utf-8 -*-
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4
#
# MDAnalysis --- https://www.mdanalysis.org
# Copyright (c) 2006-2017 The MDAnalysis Development Team and contributors
# (see the file AUTHORS for the full list of names)
#
# Released under the Lesser GNU Public Licence, v2.1 or any higher version
#
# Please cite your use of MDAnalysis in published work:
#
# R. J. Gowers, M. Linke, J. Barnoud, T. J. E. Reddy, M. N. Melo, S. L. Seyler,
# D. L. Dotson, J. Domanski, S. Buchoux, I. M. Kenney, and O. Beckstein.
# MDAnalysis: A Python package for the rapid analysis of molecular dynamics
# simulations. In S. Benthall and S. Rostrup editors, Proceedings of the 15th
# Python in Science Conference, pages 102-109, Austin, TX, 2016. SciPy.
# doi: 10.25080/majora-629e541a-00e
#
# N. Michaud-Agrawal, E. J. Denning, T. B. Woolf, and O. Beckstein.
# MDAnalysis: A Toolkit for the Analysis of Molecular Dynamics Simulations.
# J. Comput. Chem. 32 (2011), 2319--2327, doi:10.1002/jcc.21787
#

"""
CRD topology parser
===================

Read a list of atoms from a CHARMM CARD coordinate file (CRD_)
to build a basic topology.  Reads atom ids (ATOMNO), atom names (TYPES),
resids (RESID), residue numbers (RESNO), residue names (RESNames), segment ids
(SEGID) and tempfactor (Weighting).

Residues are detected through a change is either resid or resname
while segments are detected according to changes in segid.

.. note::

        By default, atomtypes and masses will be guessed on Universe creation.
        This may change in release 3.0.
        See :ref:`Guessers` for more information.

.. _CRD: https://www.charmmtutorial.org/index.php/CHARMM:The_Basics


Classes
-------

.. autoclass:: CRDParser
   :members:
   :inherited-members:

"""
import numpy as np

from ..lib.util import openany, FORTRANReader
from .base import TopologyReaderBase, change_squash
from ..core.topology import Topology
from ..core.topologyattrs import (
    Atomids,
    Atomnames,
    Resids,
    Resnames,
    Resnums,
    Segids,
    Tempfactors,
)


class CRDParser(TopologyReaderBase):
    """Parse a CHARMM CARD coordinate file for topology information.

    Reads the following Attributes:
     - Atomids
     - Atomnames
     - Tempfactors
     - Resids
     - Resnames
     - Resnums
     - Segids


    .. note::

        By default, atomtypes and masses will be guessed on Universe creation.
        This may change in release 3.0.
        See :ref:`Guessers` for more information.

    .. versionchanged:: 2.8.0
       Type and mass are not longer guessed here. Until 3.0 these will still be
       set by default through through universe.guess_TopologyAttrs() API.
    """

    format = "CRD"

    def parse(self, **kwargs):
        """Create the Topology object

        Returns
        -------
        MDAnalysis Topology object

        Todo
        ----
        Could use the resnum and temp factor better
        """
        extformat = FORTRANReader(
            "2I10,2X,A8,2X,A8,3F20.10,2X,A8,2X,A8,F20.10"
        )
        stdformat = FORTRANReader("2I5,1X,A4,1X,A4,3F10.5,1X,A4,1X,A4,F10.5")

        atomids = []
        atomnames = []
        tempfactors = []
        resids = []
        resnames = []
        resnums = []
        segids = []

        with openany(self.filename) as crd:
            for linenum, line in enumerate(crd):
                # reading header
                if line.split()[0] == "*":
                    continue
                elif line.split()[-1] == "EXT" and int(line.split()[0]):
                    r = extformat
                    continue
                elif (
                    line.split()[0] == line.split()[-1]
                    and line.split()[0] != "*"
                ):
                    r = stdformat
                    continue
                # anything else should be an atom
                try:
                    (
                        serial,
                        resnum,
                        resName,
                        name,
                        x,
                        y,
                        z,
                        segid,
                        resid,
                        tempFactor,
                    ) = r.read(line)
                except Exception:
                    errmsg = (
                        f"Check CRD format at line {linenum + 1}: "
                        f"{line.rstrip()}"
                    )
                    raise ValueError(errmsg) from None

                atomids.append(serial)
                atomnames.append(name)
                tempfactors.append(tempFactor)
                resids.append(resid)
                resnames.append(resName)
                resnums.append(resnum)
                segids.append(segid)

        # Convert to np arrays
        atomids = np.array(atomids, dtype=np.int32)
        atomnames = np.array(atomnames, dtype=object)
        tempfactors = np.array(tempfactors, dtype=np.float32)
        resids = np.array(resids, dtype=np.int32)
        resnames = np.array(resnames, dtype=object)
        resnums = np.array(resnums, dtype=np.int32)
        segids = np.array(segids, dtype=object)

        atom_residx, (res_resids, res_resnames, res_resnums, res_segids) = (
            change_squash(
                (resids, resnames), (resids, resnames, resnums, segids)
            )
        )
        res_segidx, (seg_segids,) = change_squash((res_segids,), (res_segids,))

        top = Topology(
            len(atomids),
            len(res_resids),
            len(seg_segids),
            attrs=[
                Atomids(atomids),
                Atomnames(atomnames),
                Tempfactors(tempfactors),
                Resids(res_resids),
                Resnames(res_resnames),
                Resnums(res_resnums),
                Segids(seg_segids),
            ],
            atom_resindex=atom_residx,
            residue_segindex=res_segidx,
        )

        return top



# -*- Mode: python; tab-width: 4; indent-tabs-mode:nil; coding:utf-8 -*-
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4 fileencoding=utf-8
#
# MDAnalysis --- https://www.mdanalysis.org
# Copyright (c) 2006-2017 The MDAnalysis Development Team and contributors
# (see the file AUTHORS for the full list of names)
#
# Released under the Lesser GNU Public Licence, v2.1 or any higher version
#
# Please cite your use of MDAnalysis in published work:
#
# R. J. Gowers, M. Linke, J. Barnoud, T. J. E. Reddy, M. N. Melo, S. L. Seyler,
# D. L. Dotson, J. Domanski, S. Buchoux, I. M. Kenney, and O. Beckstein.
# MDAnalysis: A Python package for the rapid analysis of molecular dynamics
# simulations. In S. Benthall and S. Rostrup editors, Proceedings of the 15th
# Python in Science Conference, pages 102-109, Austin, TX, 2016. SciPy.
# doi: 10.25080/majora-629e541a-00e
#
# N. Michaud-Agrawal, E. J. Denning, T. B. Woolf, and O. Beckstein.
# MDAnalysis: A Toolkit for the Analysis of Molecular Dynamics Simulations.
# J. Comput. Chem. 32 (2011), 2319--2327, doi:10.1002/jcc.21787
#
import pytest
import MDAnalysis as mda

from numpy.testing import assert_allclose, assert_equal

from MDAnalysisTests.coordinates.base import _SingleFrameReader
from MDAnalysisTests.coordinates.reference import RefAdKSmall
from MDAnalysis.converters.ParmEd import ParmEdConverter

from MDAnalysisTests.datafiles import (
    GRO,
    PDB,
    PSF,
    PSF_NAMD,
    PSF_cmap,
    PDB_small,
    PRM,
    PRM12,
    PRM_UreyBradley,
)


pmd = pytest.importorskip("parmed")


class TestParmEdReaderGRO:
    ref_filename = GRO

    universe = mda.Universe(pmd.load_file(GRO))
    ref = mda.Universe(GRO)

    def test_dimensions(self):
        assert_allclose(
            self.universe.trajectory.ts.dimensions,
            self.ref.trajectory.ts.dimensions,
            rtol=0,
            atol=1e-3,
            err_msg=(
                "ParmEdReader failed to get unitcell dimensions " "from ParmEd"
            ),
        )

    def test_coordinates(self):
        up = self.universe.atoms.positions
        rp = self.ref.atoms.positions
        assert_allclose(up, rp, rtol=0, atol=1e-3)


class BaseTestParmEdReader(_SingleFrameReader):
    def setUp(self):
        self.universe = mda.Universe(pmd.load_file(self.ref_filename))
        self.ref = mda.Universe(self.ref_filename)

    def test_dimensions(self):
        assert_allclose(
            self.universe.trajectory.ts.dimensions,
            self.ref.trajectory.ts.dimensions,
            rtol=0,
            atol=1e-3,
            err_msg=(
                "ParmEdReader failed to get unitcell dimensions " "from ParmEd"
            ),
        )

    def test_coordinates(self):
        up = self.universe.atoms.positions
        rp = self.ref.atoms.positions
        assert_allclose(up, rp, rtol=0, atol=1e-3)


class TestParmEdReaderPDB(BaseTestParmEdReader):

    ref_filename = RefAdKSmall.filename

    def test_uses_ParmEdReader(self):
        from MDAnalysis.coordinates.ParmEd import ParmEdReader

        assert isinstance(
            self.universe.trajectory, ParmEdReader
        ), "failed to choose ParmEdReader"


def _parmed_param_eq(a, b):
    a_idx = [a.atom1.idx, a.atom2.idx]
    b_idx = [b.atom1.idx, b.atom2.idx]

    for i in (3, 4, 5):
        atom = "atom{}".format(i)
        if hasattr(a, atom):
            if not hasattr(b, atom):
                return False
            a_idx.append(getattr(a, atom).idx)
            b_idx.append(getattr(b, atom).idx)

    atoms = a_idx == b_idx or a_idx == b_idx[::-1]
    return atoms and a.type == b.type


class BaseTestParmEdConverter:

    equal_atom_attrs = ("name", "altloc")
    almost_equal_atom_attrs = ("mass", "charge", "occupancy")
    expected_attrs = (
        "atoms",
        "bonds",
        "angles",
        "dihedrals",
        "impropers",
        "cmaps",
        "urey_bradleys",
    )

    @pytest.fixture(scope="class")
    def ref(self):
        # skip_bonds controls whether to search for bonds if it's not in the file
        return pmd.load_file(self.ref_filename, skip_bonds=True)

    @pytest.fixture(scope="class")
    def universe(self, ref):
        return mda.Universe(self.ref_filename)

    @pytest.fixture(scope="class")
    def output(self, universe):
        return universe.atoms.convert_to("PARMED")

    @pytest.fixture(scope="class")
    def roundtrip(self, ref):
        u = mda.Universe(ref)
        return u.atoms.convert_to("PARMED")

    def test_equivalent_connectivity_counts(self, universe, output):
        for attr in (
            "atoms",
            "bonds",
            "angles",
            "dihedrals",
            "impropers",
            "cmaps",
            "urey_bradleys",
        ):
            u = getattr(universe, attr, [])
            o = getattr(output, attr)
            assert len(u) == len(o)

    @pytest.mark.parametrize(
        "attr", ("bonds", "angles", "dihedrals", "impropers", "cmaps")
    )
    def test_equivalent_connectivity_values(self, universe, output, attr):
        u = getattr(universe._topology, attr, [])
        vals = u.values if u else []
        o = getattr(output, attr)
        for param in o:
            ix = [param.atom1.idx, param.atom2.idx]
            try:
                ix.append(param.atom3.idx)
            except:
                pass
            try:
                ix.append(param.atom4.idx)
            except:
                pass
            try:
                ix.append(param.atom5.idx)
            except:
                pass
            ix = tuple(ix)
            assert ix in vals or ix[::-1] in vals

    def test_equivalent_ureybradley_values(self, universe, output):
        try:
            vals = universe._topology.ureybradleys.values
        except AttributeError:
            vals = []
        o = output.urey_bradleys
        for param in o:
            ix = (param.atom1.idx, param.atom2.idx)
            assert ix in vals or ix[::-1] in vals

    def test_equivalent_atoms(self, ref, output):
        for r, o in zip(ref.atoms, output.atoms):
            for attr in self.equal_atom_attrs:
                ra = getattr(r, attr)
                oa = getattr(o, attr)
                assert (
                    ra == oa
                ), "atom {} not equal for atoms {} and {}".format(attr, r, o)

            for attr in self.almost_equal_atom_attrs:
                ra = getattr(r, attr)
                oa = getattr(o, attr)
                assert_allclose(
                    ra,
                    oa,
                    rtol=0,
                    atol=1e-2,
                    err_msg=(
                        f"atom {attr} not almost equal for atoms "
                        f"{r} and {o}"
                    ),
                )

    @pytest.mark.parametrize("attr", ("bonds", "angles", "impropers", "cmaps"))
    def test_equivalent_connectivity_types(self, ref, roundtrip, attr):
        original = getattr(ref, attr)
        for p in getattr(roundtrip, attr):
            for q in original:
                _parmed_param_eq(p, q)
            assert any(_parmed_param_eq(p, q) for q in original)

    def test_equivalent_dihedrals(self, ref, roundtrip):
        original = ref.dihedrals
        for p in roundtrip.dihedrals:
            assert any(
                (
                    _parmed_param_eq(p, q)
                    and p.improper == q.improper
                    and p.ignore_end == q.ignore_end
                )
                for q in original
            )

    def test_missing_attr(self):
        n_atoms = 10
        u = mda.Universe.empty(n_atoms)
        u.add_TopologyAttr("resid", [1])
        u.add_TopologyAttr("segid", ["DUM"])
        u.add_TopologyAttr("mass", [1] * n_atoms)
        with pytest.warns(
            UserWarning,
            match="Supplied AtomGroup was missing the following " "attributes",
        ):
            # should miss names and resnames
            u.atoms.convert_to("PARMED")


class BaseTestParmEdConverterSubset(BaseTestParmEdConverter):

    start_i = 0
    end_i = 0
    skip_i = 1

    @pytest.fixture(scope="class")
    def ref(self):
        # skip_bonds controls whether to search for bonds if it's not in the file
        struct = pmd.load_file(self.ref_filename, skip_bonds=True)
        return struct[self.start_i : self.end_i : self.skip_i]

    @pytest.fixture(scope="class")
    def universe(self):
        u = mda.Universe(self.ref_filename)
        return mda.Merge(u.atoms[self.start_i : self.end_i : self.skip_i])


class BaseTestParmEdConverterFromParmed(BaseTestParmEdConverter):

    equal_atom_attrs = ("name", "number", "altloc")

    @pytest.fixture(scope="class")
    def universe(self, ref):
        return mda.Universe(ref)

    def test_equivalent_connectivity_counts(self, ref, output):
        for attr in (
            "atoms",
            "bonds",
            "angles",
            "dihedrals",
            "impropers",
            "cmaps",
            "urey_bradleys",
        ):
            r = getattr(ref, attr)
            o = getattr(output, attr)
            assert len(r) == len(o)


class TestParmEdConverterPRM(BaseTestParmEdConverter):
    ref_filename = PRM


class TestParmEdConverterParmedPRM(BaseTestParmEdConverterFromParmed):
    ref_filename = PRM_UreyBradley


# class TestParmEdConverterPDBSubset(BaseTestParmEdConverterSubset):
#     ref_filename = PDB
#     start_i = 101
#     end_i = 8023
#     skip_i = 3

# TODO: Add Subset test for PRMs when mda.Merge accepts Universes without positions


class TestParmEdConverterParmedPSF(BaseTestParmEdConverterFromParmed):
    ref_filename = PSF_cmap


class TestParmEdConverterPSF(BaseTestParmEdConverter):
    ref_filename = PSF_NAMD


class TestParmEdConverterGROSubset(BaseTestParmEdConverterSubset):
    ref_filename = GRO
    start_i = 5
    end_i = 100


# TODO: Add Subset test for PRMs when mda.Merge accepts Universes without positions


class TestParmEdConverterPDB(BaseTestParmEdConverter):
    ref_filename = PDB_small

    # Neither MDAnalysis nor ParmEd read the mass column
    # of PDBs and are liable to guess wrong
    almost_equal_atom_attrs = ("charge", "occupancy")

    def test_equivalent_coordinates(self, ref, output):
        assert_allclose(ref.coordinates, output.coordinates, rtol=0, atol=1e-3)


def test_incorrect_object_passed_typeerror():
    err = "No atoms found in obj argument"
    with pytest.raises(TypeError, match=err):
        c = ParmEdConverter()
        c.convert("we still don't support emojis :(")


def test_old_import_warning():
    wmsg = "Please import the ParmEd classes from MDAnalysis.converters"
    with pytest.warns(DeprecationWarning, match=wmsg):
        from MDAnalysis.coordinates.ParmEd import ParmEdConverter



# -*- Mode: python; tab-width: 4; indent-tabs-mode:nil; coding:utf-8 -*-
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4
#
# MDAnalysis --- https://www.mdanalysis.org
# Copyright (c) 2006-2017 The MDAnalysis Development Team and contributors
# (see the file AUTHORS for the full list of names)
#
# Released under the Lesser GNU Public Licence, v2.1 or any higher version
#
# Please cite your use of MDAnalysis in published work:
#
# R. J. Gowers, M. Linke, J. Barnoud, T. J. E. Reddy, M. N. Melo, S. L. Seyler,
# D. L. Dotson, J. Domanski, S. Buchoux, I. M. Kenney, and O. Beckstein.
# MDAnalysis: A Python package for the rapid analysis of molecular dynamics
# simulations. In S. Benthall and S. Rostrup editors, Proceedings of the 15th
# Python in Science Conference, pages 102-109, Austin, TX, 2016. SciPy.
# doi: 10.25080/majora-629e541a-00e
#
# N. Michaud-Agrawal, E. J. Denning, T. B. Woolf, and O. Beckstein.
# MDAnalysis: A Toolkit for the Analysis of Molecular Dynamics Simulations.
# J. Comput. Chem. 32 (2011), 2319--2327, doi:10.1002/jcc.21787
#

"""Diffusion map --- :mod:`MDAnalysis.analysis.diffusionmap`
=====================================================================

:Authors: Eugen Hruska, John Detlefs
:Year: 2016
:Copyright: Lesser GNU Public License v2.1+

This module contains the non-linear dimension reduction method diffusion map.
The eigenvectors of a diffusion matrix represent the 'collective coordinates'
of a molecule; the largest eigenvalues are the more dominant collective
coordinates. Assigning physical meaning to the 'collective coordinates' is a
fundamentally difficult problem. The time complexity of the diffusion map is
:math:`O(N^3)`, where N is the number of frames in the trajectory, and the in-memory
storage complexity is :math:`O(N^2)`. Instead of a single trajectory a sample of
protein structures can be used. The sample should be equilibrated, at least
locally. The order of the sampled structures in the trajectory is irrelevant.

The :ref:`Diffusion-Map-tutorial` shows how to use diffusion map for dimension
reduction.

More details about diffusion maps are in
:footcite:p:`deLaPorte2008,Coifman-Lafon,Ferguson2011,Clementi2011`.

.. _Diffusion-Map-tutorial:

Diffusion Map tutorial
----------------------

The example uses files provided as part of the MDAnalysis test suite
(in the variables :data:`~MDAnalysis.tests.datafiles.PSF` and
:data:`~MDAnalysis.tests.datafiles.DCD`). This tutorial shows how to use the
Diffusion Map class.

First load all modules and test data

.. code-block:: python

   import MDAnalysis as mda
   import MDAnalysis.analysis.diffusionmap as diffusionmap
   from MDAnalysis.tests.datafiles import PSF, DCD

Given a universe or atom group, we can create and eigenvalue decompose
the Diffusion Matrix from that trajectory using :class:`DiffusionMap` and get
the corresponding eigenvalues and eigenvectors.

.. code-block:: python

   u = mda.Universe(PSF,DCD)

We leave determination of the appropriate scale parameter epsilon to the user,
:footcite:p:`Clementi2011` uses a complex method involving the k-nearest-neighbors
of a trajectory frame, whereas others simple use a trial-and-error approach
with a constant epsilon. Currently, the constant epsilon method is implemented
by MDAnalysis.

.. code-block:: python

   dmap = diffusionmap.DiffusionMap(u, select='backbone', epsilon=2)
   dmap.run()

From here we can perform an embedding onto the k dominant eigenvectors. The
non-linearity of the map means there is no explicit relationship between the
lower dimensional space and our original trajectory. However, this is an
isometry (distance preserving map), which means that points close in the lower
dimensional space are close in the higher-dimensional space and vice versa.
In order to embed into the most relevant low-dimensional space, there should
exist some number of dominant eigenvectors, whose corresponding eigenvalues
diminish at a constant rate until falling off, this is referred to as a
spectral gap and should be somewhat apparent for a system at equilibrium with a
high number of frames.

.. code-block:: python

   import matplotlib.pyplot as plt
   f, ax = plt.subplots()
   upper_limit = # some reasonably high number less than the n_eigenvectors
   ax.plot(dmap.eigenvalues[:upper_limit])
   ax.set(xlabel ='eigenvalue index', ylabel='eigenvalue')
   plt.tight_layout()

From here we can transform into the diffusion space

.. code-block:: python

   num_eigenvectors = # some number less than the number of frames after
   # inspecting for the spectral gap
   fit = dmap.transform(num_eigenvectors, time=1)

It can be difficult to interpret the data, and is left as a task
for the user. The `diffusion distance` between frames i and j is best
approximated by the euclidean distance  between rows i and j of
self.diffusion_space.

Classes
-------

.. autoclass:: DiffusionMap
.. autoclass:: DistanceMatrix

References
----------

If you use this Dimension Reduction method in a publication, please
cite :footcite:p:`Coifman-Lafon`.

If you choose the default metric, this module uses the fast QCP algorithm
:footcite:p:`Theobald2005` to calculate the root mean square distance (RMSD)
between two coordinate sets (as implemented in
:func:`MDAnalysis.lib.qcprot.CalcRMSDRotationalMatrix`).  When using this
module in published work please :footcite:p:`Theobald2005`.


.. footbibliography::
"""
import logging
import warnings

import numpy as np

from MDAnalysis.core.universe import Universe
from MDAnalysis.core.groups import AtomGroup, UpdatingAtomGroup
from .rms import rmsd
from .base import AnalysisBase

logger = logging.getLogger("MDAnalysis.analysis.diffusionmap")


class DistanceMatrix(AnalysisBase):
    """Calculate the pairwise distance between each frame in a trajectory
    using a given metric

    A distance matrix can be initialized on its own and used as an
    initialization argument in :class:`DiffusionMap`.

    Parameters
    ----------
    universe : `~MDAnalysis.core.universe.Universe` or `~MDAnalysis.core.groups.AtomGroup`
        The MD Trajectory for dimension reduction, remember that
        computational cost of eigenvalue decomposition
        scales at O(N^3) where N is the number of frames.
        Cost can be reduced by increasing step interval or specifying a
        start and stop value when calling :meth:`DistanceMatrix.run`.
    select : str, optional
        Any valid selection string for
        :meth:`~MDAnalysis.core.groups.AtomGroup.select_atoms`
        This selection of atoms is used to calculate the RMSD between
        different frames. Water should be excluded.
    metric : function, optional
        Maps two numpy arrays to a float, is positive definite and
        symmetric. The API for a metric requires that the arrays must have
        equal length, and that the function should have weights as an
        optional argument. Weights give each index value its own weight for
        the metric calculation over the entire arrays. Default: metric is
        set to rms.rmsd().
    cutoff : float, optional
        Specify a given cutoff for metric values to be considered equal,
        Default: 1EO-5
    weights : array, optional
        Weights to be given to coordinates for metric calculation
    verbose : bool, optional
            Show detailed progress of the calculation if set to ``True``; the
            default is ``False``.

    Attributes
    ----------
    atoms : `~MDAnalysis.core.groups.AtomGroup`
        Selected atoms in trajectory subject to dimension reduction
    results.dist_matrix : numpy.ndarray, (n_frames, n_frames)
        Array of all possible ij metric distances between frames in trajectory.
        This matrix is symmetric with zeros on the diagonal.

        .. versionadded:: 2.0.0

    dist_matrix : numpy.ndarray, (n_frames, n_frames)

        .. deprecated:: 2.0.0
                 Will be removed in MDAnalysis 3.0.0. Please use
                 :attr:`results.dist_matrix` instead.

    Example
    -------
    Often, a custom distance matrix could be useful for local
    epsilon determination or other manipulations on the diffusion
    map method. The :class:`DistanceMatrix` exists in
    :mod:`~MDAnalysis.analysis.diffusionmap` and can be passed
    as an initialization argument for :class:`DiffusionMap`.

    .. code-block:: python

        import MDAnalysis as mda
        import MDAnalysis.analysis.diffusionmap as diffusionmap
        from MDAnalysis.tests.datafiles import PSF, DCD

    Now create the distance matrix and pass it as an argument to
    :class:`DiffusionMap`.

        u = mda.Universe(PSF,DCD)
        dist_matrix = diffusionmap.DistanceMatrix(u, select='all')
        dist_matrix.run()
        dmap = diffusionmap.DiffusionMap(dist_matrix)
        dmap.run()

    .. versionchanged:: 1.0.0
       ``save()`` method has been removed. You can use ``np.save()`` on
       :attr:`DistanceMatrix.results.dist_matrix` instead.
    .. versionchanged:: 2.0.0
         :attr:`dist_matrix` is now stored in a
         :class:`MDAnalysis.analysis.base.Results` instance.
    .. versionchanged:: 2.2.0
         :class:`DistanceMatrix` now also accepts `AtomGroup`.
    .. versionchanged:: 2.8.0
         :class:`DistanceMatrix` is now correctly works with `frames=...`
         parameter (#4432) by iterating over `self._sliced_trajectory`
    """

    def __init__(
        self,
        universe,
        select="all",
        metric=rmsd,
        cutoff=1e0 - 5,
        weights=None,
        **kwargs,
    ):
        # remember that this must be called before referencing self.n_frames
        super(DistanceMatrix, self).__init__(
            universe.universe.trajectory, **kwargs
        )

        if isinstance(universe, UpdatingAtomGroup):
            wmsg = (
                "U must be a static AtomGroup. Parsing an updating AtomGroup "
                "will result in a static AtomGroup with the current frame "
                "atom selection."
            )
            warnings.warn(wmsg)

        self.atoms = universe.select_atoms(select)
        self._metric = metric
        self._cutoff = cutoff
        self._weights = weights
        self._calculated = False

    def _prepare(self):
        self.results.dist_matrix = np.zeros((self.n_frames, self.n_frames))

    def _single_frame(self):
        iframe = self._frame_index
        i_ref = self.atoms.positions
        # diagonal entries need not be calculated due to metric(x,x) == 0 in
        # theory, _ts not updated properly. Possible savings by setting a
        # cutoff for significant decimal places to sparsify matrix
        for j, ts in enumerate(self._sliced_trajectory[iframe:]):
            self._ts = ts
            j_ref = self.atoms.positions
            dist = self._metric(i_ref, j_ref, weights=self._weights)
            self.results.dist_matrix[
                self._frame_index, j + self._frame_index
            ] = (dist if dist > self._cutoff else 0)
            self.results.dist_matrix[
                j + self._frame_index, self._frame_index
            ] = self.results.dist_matrix[
                self._frame_index, j + self._frame_index
            ]
        self._ts = self._sliced_trajectory[iframe]

    @property
    def dist_matrix(self):
        wmsg = (
            "The `dist_matrix` attribute was deprecated in "
            "MDAnalysis 2.0.0 and will be removed in MDAnalysis 3.0.0. "
            "Please use `results.dist_matrix` instead."
        )
        warnings.warn(wmsg, DeprecationWarning)
        return self.results.dist_matrix

    def _conclude(self):
        self._calculated = True


class DiffusionMap(object):
    """Non-linear dimension reduction method

    Dimension reduction with diffusion mapping of selected structures in a
    trajectory.

    Attributes
    ----------
    eigenvalues: array (n_frames,)
        Eigenvalues of the diffusion map

    Methods
    -------
    run()
        Constructs an anisotropic diffusion kernel and performs eigenvalue
        decomposition on it.
    transform(n_eigenvectors, time)
        Perform an embedding of a frame into the eigenvectors representing
        the collective coordinates.


    .. versionchanged:: 2.2.0
         :class:`DiffusionMap` now also accepts `AtomGroup`.
    """

    def __init__(self, u, epsilon=1, **kwargs):
        """
        Parameters
        -------------
        u : MDAnalysis Universe or AtomGroup or DistanceMatrix object.
            Can be a Universe or AtomGroup, in which case one must supply kwargs for the
            initialization of a DistanceMatrix. Otherwise, this can be a
            DistanceMatrix already initialized. Either way, this will be made
            into a diffusion kernel.
        epsilon : Float
            Specifies the method used for the choice of scale parameter in the
            diffusion map. More information in
            :footcite:p:`Coifman-Lafon,Ferguson2011,Clementi2011`, Default: 1.
        **kwargs
            Parameters to be passed for the initialization of a
            :class:`DistanceMatrix`.
        """
        if isinstance(u, AtomGroup) or isinstance(u, Universe):
            self._dist_matrix = DistanceMatrix(u, **kwargs)
        elif isinstance(u, DistanceMatrix):
            self._dist_matrix = u
        else:
            raise ValueError(
                "U is not a Universe or AtomGroup or DistanceMatrix and"
                " so the DiffusionMap has no data to work with."
            )
        self._epsilon = epsilon

    def run(self, start=None, stop=None, step=None):
        """Create and decompose the diffusion matrix in preparation
        for a diffusion map.

        Parameters
        ----------
        start : int, optional
            start frame of analysis
        stop : int, optional
            stop frame of analysis
        step : int, optional
            number of frames to skip between each analysed frame

        .. versionchanged:: 0.19.0
           Added start/stop/step kwargs
        """
        # run only if distance matrix not already calculated
        if not self._dist_matrix._calculated:
            self._dist_matrix.run(start=start, stop=stop, step=step)
        # important for transform function and length of .run() method
        self._n_frames = self._dist_matrix.n_frames
        if self._n_frames > 5000:
            warnings.warn(
                "The distance matrix is very large, and can "
                "be very slow to compute. Consider picking a larger "
                "step size in distance matrix initialization."
            )
        self._scaled_matrix = (
            self._dist_matrix.results.dist_matrix**2 / self._epsilon
        )
        # take negative exponent of scaled matrix to create Isotropic kernel
        self._kernel = np.exp(-self._scaled_matrix)
        D_inv = np.diag(1 / self._kernel.sum(1))
        self._diff = np.dot(D_inv, self._kernel)
        self._eigenvals, self._eigenvectors = np.linalg.eig(self._diff)
        sort_idx = np.argsort(self._eigenvals)[::-1]
        self.eigenvalues = self._eigenvals[sort_idx]
        self._eigenvectors = self._eigenvectors[sort_idx]
        self._calculated = True
        return self

    def transform(self, n_eigenvectors, time):
        """Embeds a trajectory via the diffusion map

        Parameters
        ---------
        n_eigenvectors : int
            The number of dominant eigenvectors to be used for
            diffusion mapping
        time : float
            Exponent that eigenvalues are raised to for embedding, for large
            values, more dominant eigenvectors determine diffusion distance.
        Return
        ------
        diffusion_space : array (n_frames, n_eigenvectors)
            The diffusion map embedding as defined by :footcite:p:`Ferguson2011`.
        """
        return self._eigenvectors[1 : n_eigenvectors + 1,].T * (
            self.eigenvalues[1 : n_eigenvectors + 1] ** time
        )



# -*- Mode: python; tab-width: 4; indent-tabs-mode:nil; coding:utf-8 -*-
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4
#
# MDAnalysis --- https://www.mdanalysis.org
# Copyright (c) 2006-2017 The MDAnalysis Development Team and contributors
# (see the file AUTHORS for the full list of names)
#
# Released under the Lesser GNU Public Licence, v2.1 or any higher version
#
# Please cite your use of MDAnalysis in published work:
#
# R. J. Gowers, M. Linke, J. Barnoud, T. J. E. Reddy, M. N. Melo, S. L. Seyler,
# D. L. Dotson, J. Domanski, S. Buchoux, I. M. Kenney, and O. Beckstein.
# MDAnalysis: A Python package for the rapid analysis of molecular dynamics
# simulations. In S. Benthall and S. Rostrup editors, Proceedings of the 15th
# Python in Science Conference, pages 102-109, Austin, TX, 2016. SciPy.
# doi: 10.25080/majora-629e541a-00e
#
# N. Michaud-Agrawal, E. J. Denning, T. B. Woolf, and O. Beckstein.
# MDAnalysis: A Toolkit for the Analysis of Molecular Dynamics Simulations.
# J. Comput. Chem. 32 (2011), 2319--2327, doi:10.1002/jcc.21787
#


"""
=========================
Test cases for MDAnalysis
=========================

Test cases are stored in a separate packacge (:mod:`MDAnalysisTests`)
that has to be installed separately from
https://www.mdanalysis.org. For more details on the tests see
https://github.com/MDAnalysis/mdanalysis/wiki/UnitTests .

We are using the NumPy_ testing frame work; thus, numpy *must* be
installed for the tests to run at all.

Run all the tests with

.. code-block:: bash

   pytest --pyargs MDAnalysisTests



Data
====

The simulation data used in some tests are from [Beckstein2009]_ (``adk.psf``,
``adk_dims.dcd``) or unpublished simulations (O. Beckstein).

   adk_dims
      Trajectory of a macromolecular transition of the enzyme adenylate kinase
      between a closed and an open conformation. The simulation was run in
      Charmm_ c35a1.

   adk_oplsaa
      Ten frames from the first 1 ns of a equilibrium trajectory of AdK in
      water with Na+ counter ions. The OPLS/AA forcefield is used with the
      TIP4P water model. The simulation was run with Gromacs_ 4.0.2.


.. [Beckstein2009] O. Beckstein, E.J. Denning, J.R. Perilla and
   T.B. Woolf, Zipping and Unzipping of Adenylate Kinase: Atomistic
   Insights into the Ensemble of Open ↔ Closed Transitions. J Mol Biol
   394 (2009), 160–176, doi:10.1016/j.jmb.2009.09.009


Writing test cases
==================

The unittests use the :mod:`pytest_` module. See the
examples provided alongside the ``MDAnalysisTests`` module.

The `SciPy testing guidelines`_ are also a good howto for writing test cases.


.. _pytest:
   https://docs.pytest.org/en/latest/
.. _SciPy testing guidelines:
   http://projects.scipy.org/numpy/wiki/TestingGuidelines#id11
.. _Charmm: http://www.charmm.org
.. _Gromacs: http://www.gromacs.org

"""



# -*- Mode: python; tab-width: 4; indent-tabs-mode:nil; coding:utf-8 -*-
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4
#
# MDAnalysis --- https://www.mdanalysis.org
# Copyright (c) 2006-2017 The MDAnalysis Development Team and contributors
# (see the file AUTHORS for the full list of names)
#
# Released under the Lesser GNU Public Licence, v2.1 or any higher version
#
# Please cite your use of MDAnalysis in published work:
#
# R. J. Gowers, M. Linke, J. Barnoud, T. J. E. Reddy, M. N. Melo, S. L. Seyler,
# D. L. Dotson, J. Domanski, S. Buchoux, I. M. Kenney, and O. Beckstein.
# MDAnalysis: A Python package for the rapid analysis of molecular dynamics
# simulations. In S. Benthall and S. Rostrup editors, Proceedings of the 15th
# Python in Science Conference, pages 102-109, Austin, TX, 2016. SciPy.
# doi: 10.25080/majora-629e541a-00e
#
# N. Michaud-Agrawal, E. J. Denning, T. B. Woolf, and O. Beckstein.
# MDAnalysis: A Toolkit for the Analysis of Molecular Dynamics Simulations.
# J. Comput. Chem. 32 (2011), 2319--2327, doi:10.1002/jcc.21787
#

"""
Jmol selections
=================

Write :class:`MDAnalysis.core.groups.AtomGroup` selection to a `str` file
that defines a `Jmol selection`_. To be used in Jmol_ like this::

  script macro.spt
  select ~selection

The selection is named *mdanalysis001*.TODO

.. autoclass:: SelectionWriter
   :inherited-members:

.. _Jmol: http://wiki.jmol.org/index.php/Main_Page
.. _Jmol selection: http://chemapps.stolaf.edu/jmol/docs/#define
"""
from . import base


class SelectionWriter(base.SelectionWriterBase):
    format = ["Jmol", "spt"]
    ext = "spt"
    default_numterms = None
    commentfmt = "#"

    def _translate(self, atoms, **kwargs):
        # Jmol indexing is 0 based when using atom bitsets
        def _index(atom):
            return str(atom.index)

        return base.join(atoms, " ", _index)

    def _write_head(self, out, **kwargs):
        out.write("@~{name!s} ({{".format(**kwargs))

    def _write_tail(self, out, **kwargs):
        out.write("});")



# -*- Mode: python; tab-width: 4; indent-tabs-mode:nil; coding:utf-8 -*-
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4
#
# MDAnalysis --- https://www.mdanalysis.org
# Copyright (c) 2006-2017 The MDAnalysis Development Team and contributors
# (see the file AUTHORS for the full list of names)
#
# Released under the Lesser GNU Public Licence, v2.1 or any higher version
#
# Please cite your use of MDAnalysis in published work:
#
# R. J. Gowers, M. Linke, J. Barnoud, T. J. E. Reddy, M. N. Melo, S. L. Seyler,
# D. L. Dotson, J. Domanski, S. Buchoux, I. M. Kenney, and O. Beckstein.
# MDAnalysis: A Python package for the rapid analysis of molecular dynamics
# simulations. In S. Benthall and S. Rostrup editors, Proceedings of the 15th
# Python in Science Conference, pages 102-109, Austin, TX, 2016. SciPy.
# doi: 10.25080/majora-629e541a-00e
#
# N. Michaud-Agrawal, E. J. Denning, T. B. Woolf, and O. Beckstein.
# MDAnalysis: A Toolkit for the Analysis of Molecular Dynamics Simulations.
# J. Comput. Chem. 32 (2011), 2319--2327, doi:10.1002/jcc.21787
#
# TRZ Reader written by Richard J. Gowers (2013)

"""TRZ trajectory I/O  --- :mod:`MDAnalysis.coordinates.TRZ`
============================================================

Classes to read `IBIsCO`_ / `YASP`_ TRZ binary trajectories, including
coordinates, velocities and more (see attributes of the :class:`Timestep`).

Data are read and written in binary representation but because this depends on
the machine hardware architecture, MDAnalysis *always* reads and writes TRZ
trajectories in *little-endian* byte order.

.. _IBIsCO: https://onlinelibrary.wiley.com/doi/full/10.1002/jcc.21717
.. _YASP: https://www.sciencedirect.com/science/article/abs/pii/0010465593901442

Classes
-------

.. autoclass:: TRZReader
   :members:

.. autoclass:: TRZWriter
   :members:

"""
import warnings
import numpy as np
import os
import errno

from . import base
from ..exceptions import NoDataError
from .timestep import Timestep
from ..lib import util
from ..lib.util import cached, store_init_arguments
from .core import triclinic_box, triclinic_vectors


class TRZReader(base.ReaderBase):
    """Reads an IBIsCO or YASP trajectory file

    Attributes
    ----------
    ts : timestep.Timestep
         :class:`~MDAnalysis.coordinates.timestep.Timestep` object containing
         coordinates of current frame

    Note
    ----
    Binary TRZ trajectories are *always* assumed to be written in
    *little-endian* byte order and are read as such.


    .. versionchanged:: 0.11.0
       Frames now 0-based instead of 1-based.
       Extra data (Temperature, Energies, Pressures, etc) now read
       into ts.data dictionary.
       Now passes a weakref of self to ts (ts._reader).
    .. versionchanged:: 1.0.1
       Now checks for the correct `n_atoms` on reading
       and can raise :exc:`ValueError`.
    .. versionchanged:: 2.1.0
       TRZReader now returns a default :attr:`dt` of 1.0 when it cannot be
       obtained from the difference between two frames.
    .. versionchanged:: 2.3.0
       _frame attribute moved to `ts.data` dictionary.
    .. deprecated:: 2.7.0
       The TRZ Reader and Writer are deprecated as of version 2.7.0
       and will be removed in version 3.0.0.
    """

    format = "TRZ"

    units = {'time': 'ps', 'length': 'nm', 'velocity': 'nm/ps'}

    @store_init_arguments
    def __init__(self, trzfilename, n_atoms=None, **kwargs):
        """Creates a TRZ Reader

        Parameters
        ----------
        trzfilename : str
            name of input file
        n_atoms : int
            number of atoms in trajectory, must be taken from topology file!
        convert_units : bool (optional)
            converts units to MDAnalysis defaults

        Raises
        ------
        ValueError
           If `n_atoms` or the number of atoms in the topology file do not
           match the number of atoms in the trajectory.
        """
        wmsg = ("The TRZ reader is deprecated and will be removed in "
                "MDAnalysis version 3.0.0")
        warnings.warn(wmsg, DeprecationWarning)

        super(TRZReader, self).__init__(trzfilename,  **kwargs)

        if n_atoms is None:
            raise ValueError('TRZReader requires the n_atoms keyword')

        self.trzfile = util.anyopen(self.filename, 'rb')
        self._cache = dict()
        self._n_atoms = n_atoms

        self._read_trz_header()
        self.ts = Timestep(self.n_atoms,
                           velocities=True,
                           forces=self.has_force,
                           reader=self,
                           **self._ts_kwargs)

        # structured dtype of a single trajectory frame
        readarg = str(n_atoms) + '<f4'
        frame_contents = [
            ('p1', '<i4'),
            ('nframe', '<i4'),
            ('ntrj', '<i4'),
            ('natoms', '<i4'),
            ('treal', '<f8'),
            ('p2', '<2i4'),
            ('box', '<9f8'),
            ('p3', '<2i4'),
            ('pressure', '<f8'),
            ('ptensor', '<6f8'),
            ('p4', '<3i4'),
            ('etot', '<f8'),
            ('ptot', '<f8'),
            ('ek', '<f8'),
            ('T', '<f8'),
            ('p5', '<6i4'),
            ('rx', readarg),
            ('pad2', '<2i4'),
            ('ry', readarg),
            ('pad3', '<2i4'),
            ('rz', readarg),
            ('pad4', '<2i4'),
            ('vx', readarg),
            ('pad5', '<2i4'),
            ('vy', readarg),
            ('pad6', '<2i4'),
            ('vz', readarg)]
        if not self.has_force:
            frame_contents += [('pad7', '<i4')]
        else:
            frame_contents += [
                ('pad7', '<2i4'),
                ('fx', readarg),
                ('pad8', '<2i4'),
                ('fy', readarg),
                ('pad9', '<2i4'),
                ('fz', readarg),
                ('pad10', '<i4')]
        self._dtype = np.dtype(frame_contents)

        self._read_next_timestep()

    def _read_trz_header(self):
        """Reads the header of the trz trajectory"""
        self._headerdtype = np.dtype([
            ('p1', '<i4'),
            ('title', '80c'),
            ('p2', '<2i4'),
            ('force', '<i4'),
            ('p3', '<i4')])
        data = np.fromfile(self.trzfile, dtype=self._headerdtype, count=1)
        self.title = ''.join(c.decode('utf-8') for c in data['title'][0]).strip()
        if data['force'] == 10:
            self.has_force = False
        elif data['force'] == 20:
            self.has_force = True
        else:
            raise IOError

    def _read_next_timestep(self, ts=None):
        if ts is None:
            ts = self.ts

        try:
            data = np.fromfile(self.trzfile, dtype=self._dtype, count=1)
            if data['natoms'][0] != self.n_atoms:
                raise ValueError("Supplied n_atoms {} is incompatible "
                                 "with provided trajectory file. "
                                 "Maybe `topology` is wrong?".format(
                                                             self.n_atoms))
            ts.frame = data['nframe'][0] - 1  # 0 based for MDA
            ts.data['frame'] = data['ntrj'][0] # moved from attr to data
            ts.time = data['treal'][0]
            ts.dimensions = triclinic_box(*(data['box'].reshape(3, 3)))
            ts.data['pressure'] = data['pressure']
            ts.data['pressure_tensor'] = data['ptensor']
            ts.data['total_energy'] = data['etot']
            ts.data['potential_energy'] = data['ptot']
            ts.data['kinetic_energy'] = data['ek']
            ts.data['temperature'] = data['T']
            ts._x[:] = data['rx']
            ts._y[:] = data['ry']
            ts._z[:] = data['rz']
            ts._velocities[:, 0] = data['vx']
            ts._velocities[:, 1] = data['vy']
            ts._velocities[:, 2] = data['vz']
            if self.has_force:
                ts._forces[:, 0] = data['fx']
                ts._forces[:, 1] = data['fy']
                ts._forces[:, 2] = data['fz']
        except IndexError: # Raises indexerror if data has no data (EOF)
            raise IOError from None
        else:
            # Convert things read into MDAnalysis' native formats (nm -> angstroms)
            if self.convert_units:
                self.convert_pos_from_native(self.ts._pos)
                if self.ts.dimensions is not None:
                    self.convert_pos_from_native(self.ts.dimensions[:3])
                self.convert_velocities_from_native(self.ts._velocities)

            return ts

    @property
    def n_atoms(self):
        """Number of atoms in a frame"""
        return self._n_atoms

    @property
    @cached('n_frames')
    def n_frames(self):
        """Total number of frames in a trajectory"""
        try:
            return self._read_trz_n_frames(self.trzfile)
        except IOError:
            return 0

    def _read_trz_n_frames(self, trzfile):
        """Uses size of file and dtype information to determine how many frames exist

        .. versionchanged:: 0.9.0
           Now is based on filesize rather than reading entire file
        """
        fsize = os.fstat(trzfile.fileno()).st_size  # size of file in bytes

        if not (fsize - self._headerdtype.itemsize) % self._dtype.itemsize == 0:
            raise IOError("Trajectory has incomplete frames")  # check that division is sane

        nframes = int((fsize - self._headerdtype.itemsize) / self._dtype.itemsize)  # returns long int otherwise

        return nframes

    def _get_dt(self):
        """The amount of time between frames in ps

        Assumes that this step is constant (ie. 2 trajectories with different
        steps haven't been stitched together).
        Returns ``AttributeError`` in case of ``StopIteration``
        (which makes :attr:`dt` return 1.0).

        .. versionchanged:: 2.1.0
           Now returns an ``AttributeError`` if dt can't be obtained from the
           time difference between two frames.
        """
        curr_frame = self.ts.frame
        try:
            t0 = self.ts.time
            self.next()
            t1 = self.ts.time
            dt = t1 - t0
        except StopIteration:
            raise AttributeError
        else:
            return dt
        finally:
            self._read_frame(curr_frame)

    @property
    @cached('delta')
    def delta(self):
        """MD integration timestep"""
        return self.dt / self.skip_timestep

    @property
    @cached('skip_timestep')
    def skip_timestep(self):
        """Timesteps between trajectory frames"""
        curr_frame = self.ts.frame
        try:
            t0 = self.ts.data['frame']
            self.next()
            t1 = self.ts.data['frame']
            skip_timestep = t1 - t0
        except StopIteration:
            return 0
        else:
            return skip_timestep
        finally:
            self._read_frame(curr_frame)

    def _read_frame(self, frame):
        """Move to *frame* and fill timestep with data.

        .. versionchanged:: 0.11.0
           Frames now 0-based instead of 1-based
        """
        move = frame - self.ts.frame

        self._seek(move - 1)
        self._read_next_timestep()
        return self.ts

    def _seek(self, nframes):
        """Move *nframes* in the trajectory

        Note that this doens't read the trajectory (ts remains unchanged)

        .. versionadded:: 0.9.0
        """
        framesize = self._dtype.itemsize
        seeksize = framesize * nframes
        maxi_l = seeksize + 1

        if seeksize > maxi_l:
            # Workaround for seek not liking long ints
            # On python 3 this branch will never be used as we defined maxi_l
            # greater than seeksize.
            framesize = long(framesize)
            seeksize = framesize * nframes

            nreps = int(seeksize / maxi_l)  # number of max seeks we'll have to do
            rem = int(seeksize % maxi_l)  # amount leftover to do once max seeks done

            for _ in range(nreps):
                self.trzfile.seek(maxint, 1)
            self.trzfile.seek(rem, 1)
        else:
            seeksize = int(seeksize)

            self.trzfile.seek(seeksize, 1)

    def _reopen(self):
        self.close()
        self.open_trajectory()
        self._read_trz_header()  # Moves to start of first frame

    def open_trajectory(self):
        """Open the trajectory file"""
        if self.trzfile is not None:
            raise IOError(errno.EALREADY, 'TRZ file already opened', self.filename)
        if not os.path.exists(self.filename):
            raise IOError(errno.ENOENT, 'TRZ file not found', self.filename)

        self.trzfile = util.anyopen(self.filename, 'rb')

        #Reset ts
        ts = self.ts
        ts.frame = -1

        return self.trzfile

    def Writer(self, filename, n_atoms=None):
        if n_atoms is None:
            # guess that they want to write the whole timestep unless told otherwise?
            n_atoms = self.ts.n_atoms
        return TRZWriter(filename, n_atoms)

    def close(self):
        """Close trz file if it was open"""
        if self.trzfile is not None:
            self.trzfile.close()
            self.trzfile = None


class TRZWriter(base.WriterBase):
    """Writes a TRZ format trajectory.

    Note
    ----
    Binary TRZ trajectories are *always* written in *little-endian* byte order.


    .. deprecated:: 2.7.0
       The TRZ Reader and Writer are deprecated as of version 2.7.0
       and will be removed in version 3.0.0.
    """

    format = 'TRZ'
    multiframe = True

    units = {'time': 'ps', 'length': 'nm', 'velocity': 'nm/ps'}

    def __init__(self, filename, n_atoms, title='TRZ', convert_units=True):
        """Create a TRZWriter

        Parameters
        ----------
        filename : str
            name of output file
        n_atoms : int
            number of atoms in trajectory
        title : str (optional)
            title of the trajectory; the title must be 80 characters or
            shorter, a longer title raises a ValueError exception.
        convert_units : bool (optional)
            units are converted to the MDAnalysis base format; [``True``]
        """
        wmsg = ("The TRZ writer is deprecated and will be removed in "
                "MDAnalysis version 3.0.0")
        warnings.warn(wmsg, DeprecationWarning)

        self.filename = filename
        if n_atoms is None:
            raise ValueError("TRZWriter requires the n_atoms keyword")
        if n_atoms == 0:
            raise ValueError("TRZWriter: no atoms in output trajectory")
        self.n_atoms = n_atoms

        if len(title) > 80:
            raise ValueError("TRZWriter: 'title' must be 80 characters of shorter")

        self.convert_units = convert_units

        self.trzfile = util.anyopen(self.filename, 'wb')

        self._writeheader(title)

        floatsize = str(n_atoms) + '<f4'
        self.frameDtype = np.dtype([
            ('p1a', '<i4'),
            ('nframe', '<i4'),
            ('ntrj', '<i4'),
            ('natoms', '<i4'),
            ('treal', '<f8'),
            ('p1b', '<i4'),
            ('p2a', '<i4'),
            ('box', '<9f8'),
            ('p2b', '<i4'),
            ('p3a', '<i4'),
            ('pressure', '<f8'),
            ('ptensor', '<6f8'),
            ('p3b', '<i4'),
            ('p4a', '<i4'),
            ('six', '<i4'),
            ('etot', '<f8'),
            ('ptot', '<f8'),
            ('ek', '<f8'),
            ('T', '<f8'),
            ('blanks', '<2f8'),
            ('p4b', '<i4'),
            ('p5a', '<i4'),
            ('rx', floatsize),
            ('p5b', '<i4'),
            ('p6a', '<i4'),
            ('ry', floatsize),
            ('p6b', '<i4'),
            ('p7a', '<i4'),
            ('rz', floatsize),
            ('p7b', '<i4'),
            ('p8a', '<i4'),
            ('vx', floatsize),
            ('p8b', '<i4'),
            ('p9a', '<i4'),
            ('vy', floatsize),
            ('p9b', '<i4'),
            ('p10a', '<i4'),
            ('vz', floatsize),
            ('p10b', '<i4')])

    def _writeheader(self, title):
        hdt = np.dtype([
            ('pad1', '<i4'), ('title', '80c'), ('pad2', '<i4'),
            ('pad3', '<i4'), ('nrec', '<i4'), ('pad4', '<i4')])
        out = np.zeros((), dtype=hdt)
        out['pad1'], out['pad2'] = 80, 80
        out['title'] = title + ' ' * (80 - len(title))
        out['pad3'], out['pad4'] = 4, 4
        out['nrec'] = 10
        out.tofile(self.trzfile)

    def _write_next_frame(self, obj):
        """Write information associated with ``obj`` at current frame into trajectory

        Parameters
        ----------
        ag : AtomGroup or Universe


        .. versionchanged:: 1.0.0
           Renamed from `write_next_timestep` to `_write_next_frame`.
        .. versionchanged:: 2.0.0
           Deprecated support for Timestep argument has now been removed.
           Use AtomGroup or Universe as an input instead.
        """
        # Check size of ts is same as initial
        try:  # atomgroup?
            ts = obj.ts
        except AttributeError:  # universe?
            try:
                ts = obj.trajectory.ts
            except AttributeError:
                errmsg = "Input obj is neither an AtomGroup or Universe"
                raise TypeError(errmsg) from None

        if not obj.atoms.n_atoms == self.n_atoms:
            errmsg = "Number of atoms in ts different to initialisation"
            raise ValueError(errmsg)

        # Gather data, faking it when unavailable
        data = {}
        faked_attrs = []
        for att in ['pressure', 'pressure_tensor', 'total_energy',
                    'potential_energy', 'kinetic_energy', 'temperature']:
            try:
                data[att] = ts.data[att]
            except KeyError:
                if att == 'pressure_tensor':
                    data[att] = np.zeros(6, dtype=np.float64)
                else:
                    data[att] = 0.0
                faked_attrs.append(att)
        try:
            data['step'] = ts.data['frame']
        except KeyError:
            data['step'] = ts.frame
            faked_attrs.append('step')
        try:
            data['time'] = ts.time
        except AttributeError:
            data['time'] = float(ts.frame)
            faked_attrs.append('time')
        if faked_attrs:
            warnings.warn("Timestep didn't have the following attributes: '{0}', "
                          "these will be set to 0 in the output trajectory"
                          "".format(", ".join(faked_attrs)))

        # Convert other stuff into our format
        if ts.dimensions is not None:
            unitcell = triclinic_vectors(ts.dimensions).reshape(9)
        else:
            warnings.warn("Timestep didn't have dimensions information, "
                          "box will be written as all zero values")
            unitcell = np.zeros(9, dtype=np.float32)

        try:
            vels = ts.velocities
        except NoDataError:
            vels = np.zeros((self.n_atoms, 3), dtype=np.float32, order='F')
            warnings.warn("Timestep didn't have velocity information, "
                          "this will be set to zero in output trajectory. ")

        out = np.zeros((), dtype=self.frameDtype)
        out['p1a'], out['p1b'] = 20, 20
        out['nframe'] = ts.frame + 1  # TRZ wants 1 based
        out['ntrj'] = data['step']
        out['natoms'] = self.n_atoms
        out['treal'] = data['time']
        out['p2a'], out['p2b'] = 72, 72
        out['box'] = self.convert_pos_to_native(unitcell, inplace=False)
        out['p3a'], out['p3b'] = 56, 56
        out['pressure'] = data['pressure']
        out['ptensor'] = data['pressure_tensor']
        out['p4a'], out['p4b'] = 60, 60
        out['six'] = 6
        out['etot'] = data['total_energy']
        out['ptot'] = data['potential_energy']
        out['ek'] = data['kinetic_energy']
        out['T'] = data['temperature']
        out['blanks'] = 0.0, 0.0
        size = ts.n_atoms * 4  # size of float for vels & coords
        out['p5a'], out['p5b'] = size, size
        out['rx'] = self.convert_pos_to_native(ts._x, inplace=False)
        out['p6a'], out['p6b'] = size, size
        out['ry'] = self.convert_pos_to_native(ts._y, inplace=False)
        out['p7a'], out['p7b'] = size, size
        out['rz'] = self.convert_pos_to_native(ts._z, inplace=False)
        out['p8a'], out['p8b'] = size, size
        out['vx'] = self.convert_velocities_to_native(vels[:, 0], inplace=False)
        out['p9a'], out['p9b'] = size, size
        out['vy'] = self.convert_velocities_to_native(vels[:, 1], inplace=False)
        out['p10a'], out['p10b'] = size, size
        out['vz'] = self.convert_velocities_to_native(vels[:, 2], inplace=False)
        out.tofile(self.trzfile)

    def close(self):
        """Close if it was open"""
        if self.trzfile is None:
            return
        self.trzfile.close()
        self.trzfile = None
